import enum
from typing import Any, Dict, Optional, Protocol, runtime_checkable

import httpx


class HTTPMethod(str, enum.Enum):
    GET = "get"
    POST = "post"


class RequesterTransport(str, enum.Enum):
    HTTPX = "httpx"
    JSON = "json"


@runtime_checkable
class SyncRequester(Protocol):
    def __call__(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        ...


@runtime_checkable
class AsyncRequester(Protocol):
    async def __call__(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        ...



from __future__ import annotations

from typing import List, Mapping, Optional


class Error(Exception):
    def __init__(self, message: Optional[str] = None):
        self.message = message
        super().__init__(self.message)


class ServerNotReacheableError(Error):
    def __init__(self, address: str, message: Optional[str] = None):
        self.address = address
        self.message = message or f"Unable to connect to '{address}'."
        super().__init__(self.message)


class ServerNotResponsiveError(Error):
    def __init__(self, url: str, timeout: Optional[int] = None, message: Optional[str] = None):
        self.url = url
        self.timeout = timeout
        self.message = message or f"Unable to read from '{url}'."
        if timeout:
            self.message += f" (timeout: {timeout} sec)"
        super().__init__(self.message)


class GraphQLError(Error):
    def __init__(
        self,
        errors: List[str],
        query: Optional[str] = None,
        variables: Optional[dict] = None,
    ):
        self.query = query
        self.variables = variables
        self.errors = errors
        self.message = f"An error occured while executing the GraphQL Query {self.query}, {self.errors}"
        super().__init__(self.message)


class BranchNotFound(Error):
    def __init__(self, identifier: str, message: Optional[str] = None):
        self.identifier = identifier
        self.message = message or f"Unable to find the branch '{identifier}' in the Database."
        super().__init__(self.message)


class SchemaNotFound(Error):
    def __init__(self, identifier: str, message: Optional[str] = None):
        self.identifier = identifier
        self.message = message or f"Unable to find the schema '{identifier}'."
        super().__init__(self.message)


class NodeNotFound(Error):
    def __init__(
        self,
        branch_name: str,
        node_type: str,
        identifier: Mapping[str, List[str]],
        message: str = "Unable to find the node in the database.",
    ):
        self.branch_name = branch_name
        self.node_type = node_type
        self.identifier = identifier

        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"""
        {self.message}
        {self.branch_name} | {self.node_type} | {self.identifier}
        """


class FilterNotFound(Error):
    def __init__(
        self,
        identifier: str,
        kind: str,
        message: Optional[str] = None,
        filters: Optional[List[str]] = None,
    ):
        self.identifier = identifier
        self.kind = kind
        self.filters = filters or []
        self.message = message or f"{identifier!r} is not a valid filter for {self.kind!r} ({', '.join(self.filters)})."
        super().__init__(self.message)


class ValidationError(Error):
    def __init__(self, identifier: str, message: str):
        self.identifier = identifier
        self.message = message
        super().__init__(self.message)


class AuthenticationError(Error):
    def __init__(self, message: Optional[str] = None):
        self.message = message or "Authentication Error, unable to execute the query."
        super().__init__(self.message)



from __future__ import annotations

import re
from typing import Optional, Union

import pendulum
from pendulum.datetime import DateTime

REGEX_MAPPING = {
    "seconds": r"(\d+)(s|sec|second|seconds)",
    "minutes": r"(\d+)(m|min|minute|minutes)",
    "hours": r"(\d+)(h|hour|hours)",
}


class Timestamp:
    def __init__(self, value: Optional[Union[str, DateTime, Timestamp]] = None):
        if value and isinstance(value, DateTime):
            self.obj = value
        elif value and isinstance(value, self.__class__):
            self.obj = value.obj
        elif isinstance(value, str):
            self.obj = self._parse_string(value)
        else:
            self.obj = DateTime.now(tz="UTC")

    @classmethod
    def _parse_string(cls, value: str) -> DateTime:
        try:
            parsed_date = pendulum.parse(value)
            if isinstance(parsed_date, DateTime):
                return parsed_date
        except pendulum.parsing.exceptions.ParserError:
            pass

        params = {}
        for key, regex in REGEX_MAPPING.items():
            match = re.search(regex, value)
            if match:
                params[key] = int(match.group(1))

        if not params:
            raise ValueError(f"Invalid time format for {value}")

        return DateTime.now(tz="UTC").subtract(**params)

    def __repr__(self) -> str:
        return f"Timestamp: {self.to_string()}"

    def to_string(self) -> str:
        return self.obj.to_iso8601_string()

    def to_timestamp(self) -> str:
        return self.obj.int_timestamp

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self.obj == other.obj

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self.obj < other.obj

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self.obj > other.obj

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self.obj <= other.obj

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self.obj >= other.obj

    def add_delta(self, hours: int = 0, minutes: int = 0, seconds: int = 0) -> Timestamp:
        time = self.obj.add(hours=hours, minutes=minutes, seconds=seconds)
        return Timestamp(time)



from __future__ import annotations

import ipaddress
import re
from copy import copy
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    get_args,
)

from infrahub_sdk.exceptions import Error, FilterNotFound, NodeNotFound
from infrahub_sdk.graphql import Mutation
from infrahub_sdk.schema import GenericSchema, RelationshipCardinality, RelationshipKind
from infrahub_sdk.timestamp import Timestamp
from infrahub_sdk.utils import compare_lists, get_flat_value
from infrahub_sdk.uuidt import UUIDT

if TYPE_CHECKING:
    from typing_extensions import Self

    from infrahub_sdk.client import InfrahubClient, InfrahubClientSync
    from infrahub_sdk.schema import AttributeSchema, NodeSchema, RelationshipSchema

# pylint: disable=too-many-lines

PROPERTIES_FLAG = ["is_visible", "is_protected"]
PROPERTIES_OBJECT = ["source", "owner"]
SAFE_VALUE = re.compile(r"(^[\. /:a-zA-Z0-9_-]+$)|(^$)")

IP_TYPES = Union[
    ipaddress.IPv4Interface,
    ipaddress.IPv6Interface,
    ipaddress.IPv4Network,
    ipaddress.IPv6Network,
]


class Attribute:
    def __init__(self, name: str, schema: AttributeSchema, data: Union[Any, dict]):
        self.name = name
        self._schema = schema

        if not isinstance(data, dict):
            data = {"value": data}

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self._read_only = ["updated_at", "is_inherited"]

        self.id: Optional[str] = data.get("id", None)

        self.value: Optional[Any] = data.get("value", None)

        if self.value:
            value_mapper: Dict[str, Callable] = {
                "IPHost": ipaddress.ip_interface,
                "IPNetwork": ipaddress.ip_network,
            }
            mapper = value_mapper.get(schema.kind, lambda value: value)
            self.value = mapper(data.get("value"))

        self.is_inherited: Optional[bool] = data.get("is_inherited", None)
        self.updated_at: Optional[str] = data.get("updated_at", None)

        self.is_visible: Optional[bool] = data.get("is_visible", None)
        self.is_protected: Optional[bool] = data.get("is_protected", None)

        self.source: Optional[NodeProperty] = None
        self.owner: Optional[NodeProperty] = None

        for prop_name in self._properties_object:
            if data.get(prop_name):
                setattr(self, prop_name, NodeProperty(data=data.get(prop_name)))

    def _generate_input_data(self) -> Optional[Dict]:
        data: Dict[str, Any] = {}
        variables: Dict[str, Any] = {}

        if self.value is None:
            return data

        if isinstance(self.value, str):
            if SAFE_VALUE.match(self.value):
                data["value"] = self.value
            else:
                var_name = f"value_{UUIDT.new().hex}"
                variables[var_name] = self.value
                data["value"] = f"${var_name}"
        elif isinstance(self.value, get_args(IP_TYPES)):
            data["value"] = self.value.with_prefixlen
        else:
            data["value"] = self.value

        for prop_name in self._properties_flag:
            if getattr(self, prop_name) is not None:
                data[prop_name] = getattr(self, prop_name)

        for prop_name in self._properties_object:
            if getattr(self, prop_name) is not None:
                data[prop_name] = getattr(self, prop_name)._generate_input_data()

        return {"data": data, "variables": variables}

    def _generate_query_data(self) -> Optional[Dict]:
        data: Dict[str, Any] = {"value": None}

        for prop_name in self._properties_flag:
            data[prop_name] = None
        for prop_name in self._properties_object:
            data[prop_name] = {"id": None, "display_label": None, "__typename": None}

        return data


class RelatedNodeBase:
    def __init__(
        self,
        branch: str,
        schema: RelationshipSchema,
        data: Union[Any, dict],
        name: Optional[str] = None,
    ):
        self.schema = schema
        self.name = name

        self._branch = branch

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self._peer = None
        self._id: Optional[str] = None
        self._display_label: Optional[str] = None
        self._typename: Optional[str] = None

        if isinstance(data, (InfrahubNode, InfrahubNodeSync)):
            self._peer = data

            for prop in self._properties:
                setattr(self, prop, None)

        elif not isinstance(data, dict):
            data = {"id": data}

        if isinstance(data, dict):
            # To support both with and without pagination, we split data into node_data and properties_data
            # We should probably clean that once we'll remove the code without pagination.
            node_data = data.get("node", data)
            properties_data = data.get("properties", data)

            if node_data:
                self._id = node_data.get("id", None)
                self._display_label = node_data.get("display_label", None)
                self._typename = node_data.get("__typename", None)

            self.updated_at: Optional[str] = data.get("updated_at", data.get("_relation__updated_at", None))

            # FIXME, we won't need that once we are only supporting paginated results
            if self._typename and self._typename.startswith("Related"):
                self._typename = self._typename[7:]

            for prop in self._properties:
                prop_data = properties_data.get(prop, properties_data.get(f"_relation__{prop}", None))
                if prop_data and isinstance(prop_data, dict) and "id" in prop_data:
                    setattr(self, prop, prop_data["id"])
                elif prop_data and isinstance(prop_data, (str, bool)):
                    setattr(self, prop, prop_data)
                else:
                    setattr(self, prop, None)

    @property
    def initialized(self) -> bool:
        if self.id:
            return True
        return False

    @property
    def id(self) -> Optional[str]:
        if self._peer:
            return self._peer.id
        return self._id

    @property
    def display_label(self) -> Optional[str]:
        if self._peer:
            return self._peer.display_label
        return self._display_label

    @property
    def typename(self) -> Optional[str]:
        if self._peer:
            return self._peer.typename
        return self._typename

    def _generate_input_data(self) -> Dict[str, Any]:
        data = {}

        if self.id is not None:
            data["id"] = self.id

        for prop_name in self._properties:
            if getattr(self, prop_name) is not None:
                data[f"_relation__{prop_name}"] = getattr(self, prop_name)

        return data

    @classmethod
    def _generate_query_data(cls) -> Dict:
        data: Dict[str, Any] = {"node": {"id": None, "display_label": None, "__typename": None}}

        properties: Dict[str, Any] = {}
        for prop_name in PROPERTIES_FLAG:
            properties[prop_name] = None
        for prop_name in PROPERTIES_OBJECT:
            properties[prop_name] = {
                "id": None,
                "display_label": None,
                "__typename": None,
            }

        if properties:
            data["properties"] = properties
        return data


class RelatedNode(RelatedNodeBase):
    def __init__(
        self,
        client: InfrahubClient,
        branch: str,
        schema: RelationshipSchema,
        data: Union[Any, dict],
        name: Optional[str] = None,
    ):
        self._client = client
        super().__init__(branch=branch, schema=schema, data=data, name=name)

    async def fetch(self) -> None:
        if not self.id or not self.typename:
            raise Error("Unable to fetch the peer, id and/or typename are not defined")

        self._peer = await self._client.get(ids=[self.id], kind=self.typename, populate_store=True)

    @property
    def peer(self) -> InfrahubNode:
        return self.get()

    def get(self) -> InfrahubNode:
        if self._peer:
            return self._peer  # type: ignore[return-value]

        if not self.id:
            raise ValueError("Node id but be defined to query it.")

        if self.id and self.typename:
            return self._client.store.get(key=self.id, kind=self.typename)  # type: ignore[return-value]

        raise NodeNotFound(
            branch_name=self._branch,
            node_type=self.schema.peer,
            identifier={"key": [self.id]},
        )


class RelatedNodeSync(RelatedNodeBase):
    def __init__(
        self,
        client: InfrahubClientSync,
        branch: str,
        schema: RelationshipSchema,
        data: Union[Any, dict],
        name: Optional[str] = None,
    ):
        self._client = client
        super().__init__(branch=branch, schema=schema, data=data, name=name)

    def fetch(self) -> None:
        if not self.id or not self.typename:
            raise Error("Unable to fetch the peer, id and/or typename are not defined")

        self._peer = self._client.get(ids=[self.id], kind=self.typename, populate_store=True)

    @property
    def peer(self) -> InfrahubNodeSync:
        return self.get()

    def get(self) -> InfrahubNodeSync:
        if self._peer:
            return self._peer  # type: ignore[return-value]

        if not self.id:
            raise ValueError("Node id but be defined to query it.")

        if self.id and self.typename:
            return self._client.store.get(key=self.id, kind=self.typename)  # type: ignore[return-value]

        raise NodeNotFound(
            branch_name=self._branch,
            node_type=self.schema.peer,
            identifier={"key": [self.id]},
        )


class RelationshipManagerBase:
    def __init__(self, name: str, branch: str, schema: RelationshipSchema):
        self.name = name
        self.schema = schema
        self.branch = branch

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self.peers: List[Union[RelatedNode, RelatedNodeSync]] = []

    @property
    def peer_ids(self) -> List[str]:
        return [peer.id for peer in self.peers if peer.id]

    def _generate_input_data(self) -> List[Dict]:
        return [peer._generate_input_data() for peer in self.peers]

    @classmethod
    def _generate_query_data(cls) -> Dict:
        data: Dict[str, Any] = {
            "count": None,
            "edges": {"node": {"id": None, "display_label": None, "__typename": None}},
        }

        properties: Dict[str, Any] = {}
        for prop_name in PROPERTIES_FLAG:
            properties[prop_name] = None
        for prop_name in PROPERTIES_OBJECT:
            properties[prop_name] = {
                "id": None,
                "display_label": None,
                "__typename": None,
            }

        if properties:
            data["edges"]["properties"] = properties
        return data


class RelationshipManager(RelationshipManagerBase):
    def __init__(
        self,
        name: str,
        client: InfrahubClient,
        node: InfrahubNode,
        branch: str,
        schema: RelationshipSchema,
        data: Union[Any, dict],
    ):
        self.client = client
        self.node = node

        self.initialized = data is not None

        super().__init__(name=name, schema=schema, branch=branch)

        if data is None:
            return

        if isinstance(data, list):
            for item in data:
                self.peers.append(
                    RelatedNode(
                        name=name,
                        client=self.client,
                        branch=self.branch,
                        schema=schema,
                        data=item,
                    )
                )

        elif isinstance(data, dict) and "edges" in data:
            for item in data["edges"]:
                self.peers.append(
                    RelatedNode(
                        name=name,
                        client=self.client,
                        branch=self.branch,
                        schema=schema,
                        data=item,
                    )
                )

        else:
            raise ValueError(f"Unexpected format for {name} found a {type(data)}, {data}")

    def __getitem__(self, item: int) -> RelatedNode:
        return self.peers[item]  # type: ignore[return-value]

    async def fetch(self) -> None:
        if not self.initialized:
            exclude = self.node._schema.relationship_names + self.node._schema.attribute_names
            exclude.remove(self.schema.name)
            node = await self.client.get(
                kind=self.node._schema.kind,
                id=self.node.id,
                include=[self.schema.name],
                exclude=exclude,
            )
            rm = getattr(node, self.schema.name)
            self.peers = rm.peers
            self.initialized = True

        for peer in self.peers:
            await peer.fetch()  # type: ignore[misc]

    def add(self, data: Union[str, RelatedNode, dict]) -> None:
        """Add a new peer to this relationship."""
        new_node = RelatedNode(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if new_node.id and new_node.id not in self.peer_ids:
            self.peers.append(
                RelatedNode(
                    schema=self.schema,
                    client=self.client,
                    branch=self.branch,
                    data=data,
                )
            )

    def remove(self, data: Union[str, RelatedNode, dict]) -> None:
        node_to_remove = RelatedNode(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if node_to_remove.id and node_to_remove.id in self.peer_ids:
            idx = self.peer_ids.index(node_to_remove.id)
            if self.peers[idx].id != node_to_remove.id:
                raise IndexError(f"Unexpected situation, the node with the index {idx} should be {node_to_remove.id}")

            self.peers.pop(idx)


class RelationshipManagerSync(RelationshipManagerBase):
    def __init__(
        self,
        name: str,
        client: InfrahubClientSync,
        node: InfrahubNodeSync,
        branch: str,
        schema: RelationshipSchema,
        data: Union[Any, dict],
    ):
        self.client = client
        self.node = node

        self.initialized = data is not None

        super().__init__(name=name, schema=schema, branch=branch)

        if data is None:
            return

        if isinstance(data, list):
            for item in data:
                self.peers.append(
                    RelatedNodeSync(
                        name=name,
                        client=self.client,
                        branch=self.branch,
                        schema=schema,
                        data=item,
                    )
                )

        elif isinstance(data, dict) and "edges" in data:
            for item in data["edges"]:
                self.peers.append(
                    RelatedNodeSync(
                        name=name,
                        client=self.client,
                        branch=self.branch,
                        schema=schema,
                        data=item,
                    )
                )

        else:
            raise ValueError(f"Unexpected format for {name} found a {type(data)}, {data}")

    def __getitem__(self, item: int) -> RelatedNodeSync:
        return self.peers[item]  # type: ignore[return-value]

    def fetch(self) -> None:
        if not self.initialized:
            exclude = self.node._schema.relationship_names + self.node._schema.attribute_names
            exclude.remove(self.schema.name)
            node = self.client.get(
                kind=self.node._schema.kind,
                id=self.node.id,
                include=[self.schema.name],
                exclude=exclude,
            )
            rm = getattr(node, self.schema.name)
            self.peers = rm.peers
            self.initialized = True

        for peer in self.peers:
            peer.fetch()

    def add(self, data: Union[str, RelatedNodeSync, dict]) -> None:
        """Add a new peer to this relationship."""
        new_node = RelatedNodeSync(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if new_node.id and new_node.id not in self.peer_ids:
            self.peers.append(
                RelatedNodeSync(
                    schema=self.schema,
                    client=self.client,
                    branch=self.branch,
                    data=data,
                )
            )

    def remove(self, data: Union[str, RelatedNodeSync, dict]) -> None:
        node_to_remove = RelatedNodeSync(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if node_to_remove.id and node_to_remove.id in self.peer_ids:
            idx = self.peer_ids.index(node_to_remove.id)
            if self.peers[idx].id != node_to_remove.id:
                raise IndexError(f"Unexpected situation, the node with the index {idx} should be {node_to_remove.id}")

            self.peers.pop(idx)


class InfrahubNodeBase:
    def __init__(
        self,
        schema: Union[NodeSchema, GenericSchema],
        branch: str,
        data: Optional[dict] = None,
    ) -> None:
        self._schema = schema
        self._data = data
        self._branch = branch
        self._existing: bool = True

        extracted_uuid = data.get("id", None) if isinstance(data, dict) else None
        self.id = extracted_uuid or str(UUIDT())
        self.display_label: Optional[str] = data.get("display_label", None) if isinstance(data, dict) else None
        self.typename: Optional[str] = data.get("__typename", schema.kind) if isinstance(data, dict) else schema.kind

        self._attributes = [item.name for item in self._schema.attributes]
        self._relationships = [item.name for item in self._schema.relationships]

        if not extracted_uuid:
            self._existing = False

        self._init_attributes(data)
        self._init_relationships(data)

    def _init_attributes(self, data: Optional[dict] = None) -> None:
        for attr_name in self._attributes:
            attr_schema = [attr for attr in self._schema.attributes if attr.name == attr_name][0]
            attr_data = data.get(attr_name, None) if isinstance(data, dict) else None
            setattr(
                self,
                attr_name,
                Attribute(name=attr_name, schema=attr_schema, data=attr_data),
            )

    def _init_relationships(self, data: Optional[dict] = None) -> None:
        pass

    def __repr__(self) -> str:
        if self.display_label:
            return self.display_label
        if not self._existing:
            return f"{self._schema.kind} ({self.id})[NEW]"

        return f"{self._schema.kind} ({self.id}) "

    def _generate_input_data(self, update: bool = False) -> Dict[str, Dict]:
        """Generate a dictionnary that represent the input data required by a mutation.

        Returns:
            Dict[str, Dict]: Representation of an input data in dict format
        """
        # pylint: disable=too-many-branches
        data = {}
        variables = {}
        for item_name in self._attributes:
            attr: Attribute = getattr(self, item_name)
            attr_data = attr._generate_input_data()

            # NOTE, this code has been inherited when we splitted attributes and relationships
            # into 2 loops, most likely it's possible to simply it
            if attr_data and isinstance(attr_data, dict):
                if variable_values := attr_data.get("data"):
                    data[item_name] = variable_values
                else:
                    data[item_name] = attr_data
                if variable_names := attr_data.get("variables"):
                    variables.update(variable_names)

            elif attr_data and isinstance(attr_data, list):
                data[item_name] = attr_data

        for item_name in self._relationships:
            rel_schema = self._schema.get_relationship(name=item_name)
            if not rel_schema:
                continue
            if rel_schema.kind in [RelationshipKind.GROUP, RelationshipKind.COMPONENT]:
                continue

            rel: Union[RelatedNodeBase, RelationshipManagerBase] = getattr(self, item_name)

            # BLOCKED by https://github.com/opsmill/infrahub/issues/330
            # if (
            #     item is None
            #     and item_name in self._relationships
            #     and self._schema.get_relationship(item_name).cardinality == "one"
            # ):
            #     data[item_name] = None
            #     continue
            # el
            if rel is None:
                continue

            rel_data = rel._generate_input_data()

            if rel_data and isinstance(rel_data, dict):
                if variable_values := rel_data.get("data"):
                    data[item_name] = variable_values
                else:
                    data[item_name] = rel_data
                if variable_names := rel_data.get("variables"):
                    variables.update(variable_names)

            elif rel_data and isinstance(rel_data, list):
                data[item_name] = rel_data
            elif rel_schema.cardinality == RelationshipCardinality.MANY:
                data[item_name] = []

        if update:
            data, variables = self._strip_unmodified(data=data, variables=variables)

        mutation_variables = {key: type(value) for key, value in variables.items()}

        return {
            "data": {"data": data},
            "variables": variables,
            "mutation_variables": mutation_variables,
        }

    @staticmethod
    def _strip_unmodified_dict(data: dict, original_data: dict, variables: dict, item: str) -> None:
        for item_key in original_data[item].keys():
            if isinstance(data[item], dict):
                for property_name in PROPERTIES_OBJECT:
                    if item_key == property_name and isinstance(original_data[item][property_name], dict):
                        if original_data[item][property_name].get("id"):
                            original_data[item][property_name] = original_data[item][property_name]["id"]
                if item_key in data[item].keys():
                    if item_key == "id" and len(data[item].keys()) > 1:
                        # Related nodes typically require an ID. So the ID is only
                        # removed if it's the last key in the current context
                        continue
                    variable_key = None
                    if isinstance(data[item][item_key], str):
                        variable_key = data[item][item_key][1:]

                    if original_data[item][item_key] == data[item][item_key]:
                        data[item].pop(item_key)
                    elif variable_key in variables and original_data[item][item_key] == variables[variable_key]:
                        data[item].pop(item_key)
                        variables.pop(variable_key)

        if not data[item]:
            data.pop(item)

    def _strip_unmodified(self, data: dict, variables: dict) -> Tuple[dict, dict]:
        original_data = self._data or {}
        for relationship in self._relationships:
            relationship_property = getattr(self, relationship)
            if relationship_property and not relationship_property.initialized and relationship in data:
                data.pop(relationship)
        for item in original_data.keys():
            if item in data.keys():
                if data[item] == original_data[item]:
                    data.pop(item)
                    continue
                if isinstance(original_data[item], dict):
                    self._strip_unmodified_dict(
                        data=data,
                        original_data=original_data,
                        variables=variables,
                        item=item,
                    )
                    if item in self._relationships and original_data[item].get("node"):
                        relationship_data_cardinality_one = copy(original_data)
                        relationship_data_cardinality_one[item] = original_data[item]["node"]
                        self._strip_unmodified_dict(
                            data=data,
                            original_data=relationship_data_cardinality_one,
                            variables=variables,
                            item=item,
                        )
                        # Run again to remove the "id" key if it's the last one remaining
                        self._strip_unmodified_dict(
                            data=data,
                            original_data=relationship_data_cardinality_one,
                            variables=variables,
                            item=item,
                        )

        return data, variables

    @staticmethod
    def _strip_alias(data: dict) -> Dict[str, Dict]:
        clean = {}

        for key, value in data.items():
            if "__alias__" in key:
                clean_key = key.split("__")[-1]
                clean[clean_key] = value
            else:
                clean[key] = value

        return clean

    def generate_query_data_init(
        self,
        filters: Optional[Dict[str, Any]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ) -> Dict[str, Union[Any, Dict]]:
        data: Dict[str, Any] = {
            "count": None,
            "edges": {"node": {"id": None, "display_label": None, "__typename": None}},
        }

        data["@filters"] = filters or {}

        if offset:
            data["@filters"]["offset"] = offset

        if limit:
            data["@filters"]["limit"] = limit

        if include and exclude:
            in_both, _, _ = compare_lists(include, exclude)
            if in_both:
                raise ValueError(f"{in_both} are part of both include and exclude")

        return data

    def generate_query_data_node(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        inherited: bool = True,
        insert_alias: bool = False,
    ) -> Dict[str, Union[Any, Dict]]:
        """Generate the node part of a GraphQL Query with attributes and nodes.

        Args:
            include (Optional[List[str]], optional): List of attributes or relationships to include. Defaults to None.
            exclude (Optional[List[str]], optional): List of attributes or relationships to exclude. Defaults to None.
            inherited (bool, optional): Indicated of the attributes and the relationships inherited from generics should be included as well. Defaults to True.

        Returns:
            Dict[str, Union[Any, Dict]]: Query in Dict format
        """
        # pylint: disable=too-many-branches

        data: Dict[str, Any] = {}

        for attr_name in self._attributes:
            if exclude and attr_name in exclude:
                continue

            attr: Attribute = getattr(self, attr_name)

            if not inherited and attr._schema.inherited:
                continue

            attr_data = attr._generate_query_data()
            if attr_data:
                data[attr_name] = attr_data
                if insert_alias:
                    data[attr_name]["@alias"] = f"__alias__{self._schema.kind}__{attr_name}"
            elif insert_alias:
                if insert_alias:
                    data[attr_name] = {"@alias": f"__alias__{self._schema.kind}__{attr_name}"}

        for rel_name in self._relationships:
            if exclude and rel_name in exclude:
                continue

            rel_schema = self._schema.get_relationship(name=rel_name)

            if not rel_schema or (not inherited and rel_schema.inherited):
                continue

            if (
                rel_schema.cardinality == RelationshipCardinality.MANY  # type: ignore[union-attr]
                and rel_schema.kind not in [RelationshipKind.ATTRIBUTE, RelationshipKind.PARENT]  # type: ignore[union-attr]
                and not (include and rel_name in include)
            ):
                continue

            if rel_schema and rel_schema.cardinality == "one":
                rel_data = RelatedNode._generate_query_data()
            elif rel_schema and rel_schema.cardinality == "many":
                rel_data = RelationshipManager._generate_query_data()
            data[rel_name] = rel_data
            if insert_alias:
                data[rel_name]["@alias"] = f"__alias__{self._schema.kind}__{rel_name}"

        return data

    def validate_filters(self, filters: Optional[Dict[str, Any]] = None) -> bool:
        if not filters:
            return True

        for filter_name in filters.keys():
            found = False
            for filter_schema in self._schema.filters:
                if filter_name == filter_schema.name:
                    found = True
                    break
            if not found:
                valid_filters = [entry.name for entry in self._schema.filters]
                raise FilterNotFound(
                    identifier=filter_name,
                    kind=self._schema.kind,
                    filters=valid_filters,
                )

        return True

    def extract(self, params: Dict[str, str]) -> Dict[str, Any]:
        """Extract some datapoints defined in a flat notation."""
        result: Dict[str, Any] = {}
        for key, value in params.items():
            result[key] = get_flat_value(self, key=value)

        return result


class InfrahubNode(InfrahubNodeBase):
    def __init__(
        self,
        client: InfrahubClient,
        schema: Union[NodeSchema, GenericSchema],
        branch: Optional[str] = None,
        data: Optional[dict] = None,
    ) -> None:
        self._client = client
        self.__class__ = type(f"{schema.kind}InfrahubNode", (self.__class__,), {})

        if isinstance(data, dict) and "node" in data:
            data = data.get("node")

        super().__init__(schema=schema, branch=branch or client.default_branch, data=data)

    @classmethod
    async def from_graphql(
        cls,
        client: InfrahubClient,
        branch: str,
        data: dict,
        schema: Optional[Union[NodeSchema, GenericSchema]] = None,
    ) -> Self:
        if not schema:
            node_kind = data.get("__typename", None) or data.get("node", {}).get("__typename", None)
            if not node_kind:
                raise ValueError("Unable to determine the type of the node, __typename not present in data")
            schema = await client.schema.get(kind=node_kind)

        return cls(client=client, schema=schema, branch=branch, data=cls._strip_alias(data))

    def _init_relationships(self, data: Optional[dict] = None) -> None:
        for rel_name in self._relationships:
            rel_schema = [rel for rel in self._schema.relationships if rel.name == rel_name][0]
            rel_data = data.get(rel_name, None) if isinstance(data, dict) else None

            if rel_schema.cardinality == "one":
                setattr(self, f"_{rel_name}", None)
                setattr(
                    self.__class__,
                    rel_name,
                    generate_relationship_property(name=rel_name, node=self, node_class=RelatedNode),  # type: ignore[arg-type]
                )
                setattr(self, rel_name, rel_data)
            else:
                setattr(
                    self,
                    rel_name,
                    RelationshipManager(
                        name=rel_name,
                        client=self._client,
                        node=self,
                        branch=self._branch,
                        schema=rel_schema,
                        data=rel_data,
                    ),
                )

    async def delete(self, at: Optional[Timestamp] = None) -> None:
        at = Timestamp(at)
        input_data = {"data": {"id": self.id}}
        mutation_query = {"ok": None}
        query = Mutation(
            mutation=f"{self._schema.kind}Delete",
            input_data=input_data,
            query=mutation_query,
        )
        await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            at=at,
            tracker=f"mutation-{str(self._schema.kind).lower()}-delete",
        )

    async def save(self, at: Optional[Timestamp] = None) -> None:
        at = Timestamp(at)
        if self._existing is False:
            await self._create(at=at)
        else:
            await self._update(at=at)

        self._client.store.set(key=self.id, node=self)

    async def generate_query_data(
        self,
        filters: Optional[Dict[str, Any]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        fragment: bool = False,
    ) -> Dict[str, Union[Any, Dict]]:
        data = self.generate_query_data_init(
            filters=filters,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
        )
        data["edges"]["node"].update(self.generate_query_data_node(include=include, exclude=exclude, inherited=True))

        if isinstance(self._schema, GenericSchema) and fragment:
            for child in self._schema.used_by:
                child_schema = await self._client.schema.get(kind=child)
                child_node = InfrahubNode(client=self._client, schema=child_schema)

                # Add the attribute and the relationship already part of the parent to the exclude list for the children
                exclude_parent = self._attributes + self._relationships
                _, _, only_in_list2 = compare_lists(list1=include or [], list2=exclude_parent)

                exclude_child = only_in_list2
                if exclude:
                    exclude_child += exclude

                child_data = child_node.generate_query_data_node(
                    include=include,
                    exclude=exclude_child,
                    inherited=False,
                    insert_alias=True,
                )

                if child_data:
                    data["edges"]["node"][f"...on {child}"] = child_data

        return {self._schema.kind: data}

    async def _create(self, at: Timestamp) -> None:
        input_data = self._generate_input_data()
        input_data["data"]["data"]["id"] = self.id
        mutation_query = {"ok": None, "object": {"id": None}}
        mutation_name = f"{self._schema.kind}Create"
        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )
        await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            at=at,
            tracker=f"mutation-{str(self._schema.kind).lower()}-create",
            variables=input_data["variables"],
        )
        self._existing = True

    async def _update(self, at: Timestamp) -> None:
        input_data = self._generate_input_data(update=True)
        input_data["data"]["data"]["id"] = self.id
        mutation_query = {"ok": None, "object": {"id": None}}
        query = Mutation(
            mutation=f"{self._schema.kind}Update",
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )
        await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            at=at,
            tracker=f"mutation-{str(self._schema.kind).lower()}-update",
            variables=input_data["variables"],
        )


class InfrahubNodeSync(InfrahubNodeBase):
    def __init__(
        self,
        client: InfrahubClientSync,
        schema: Union[NodeSchema, GenericSchema],
        branch: Optional[str] = None,
        data: Optional[dict] = None,
    ) -> None:
        self.__class__ = type(f"{schema.kind}InfrahubNodeSync", (self.__class__,), {})
        self._client = client

        if isinstance(data, dict) and "node" in data:
            data = data.get("node")

        super().__init__(schema=schema, branch=branch or client.default_branch, data=data)

    @classmethod
    def from_graphql(
        cls,
        client: InfrahubClientSync,
        branch: str,
        data: dict,
        schema: Optional[Union[NodeSchema, GenericSchema]] = None,
    ) -> Self:
        if not schema:
            node_kind = data.get("__typename", None) or data.get("node", {}).get("__typename", None)
            if not node_kind:
                raise ValueError("Unable to determine the type of the node, __typename not present in data")
            schema = client.schema.get(kind=node_kind)

        return cls(client=client, schema=schema, branch=branch, data=cls._strip_alias(data))

    def _init_relationships(self, data: Optional[dict] = None) -> None:
        for rel_name in self._relationships:
            rel_schema = [rel for rel in self._schema.relationships if rel.name == rel_name][0]
            rel_data = data.get(rel_name, None) if isinstance(data, dict) else None

            if rel_schema.cardinality == "one":
                setattr(self, f"_{rel_name}", None)
                setattr(
                    self.__class__,
                    rel_name,
                    generate_relationship_property(name=rel_name, node=self, node_class=RelatedNodeSync),  # type: ignore[arg-type]
                )
                setattr(self, rel_name, rel_data)
            else:
                setattr(
                    self,
                    rel_name,
                    RelationshipManagerSync(
                        name=rel_name,
                        client=self._client,
                        node=self,
                        branch=self._branch,
                        schema=rel_schema,
                        data=rel_data,
                    ),
                )

    def delete(self, at: Optional[Timestamp] = None) -> None:
        at = Timestamp(at)
        input_data = {"data": {"id": self.id}}
        mutation_query = {"ok": None}
        query = Mutation(
            mutation=f"{self._schema.kind}Delete",
            input_data=input_data,
            query=mutation_query,
        )
        self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            at=at,
            tracker=f"mutation-{str(self._schema.kind).lower()}-delete",
        )

    def save(self, at: Optional[Timestamp] = None) -> None:
        at = Timestamp(at)
        if self._existing is False:
            self._create(at=at)
        else:
            self._update(at=at)

    def generate_query_data(
        self,
        filters: Optional[Dict[str, Any]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        fragment: bool = False,
    ) -> Dict[str, Union[Any, Dict]]:
        data = self.generate_query_data_init(
            filters=filters,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
        )
        data["edges"]["node"].update(self.generate_query_data_node(include=include, exclude=exclude, inherited=True))

        if isinstance(self._schema, GenericSchema) and fragment:
            for parent in self._schema.used_by:
                parent_schema = self._client.schema.get(kind=parent)
                parent_node = InfrahubNodeSync(client=self._client, schema=parent_schema)
                parent_data = parent_node.generate_query_data_node(
                    include=include, exclude=exclude, inherited=False, insert_alias=True
                )
                if parent_data:
                    data["edges"]["node"][f"...on {parent}"] = parent_data

        return {self._schema.kind: data}

    def _create(self, at: Timestamp) -> None:
        input_data = self._generate_input_data()
        mutation_query = {"ok": None, "object": {"id": None}}
        mutation_name = f"{self._schema.kind}Create"
        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            at=at,
            tracker=f"mutation-{str(self._schema.kind).lower()}-create",
            variables=input_data["variables"],
        )
        self._existing = True

    def _update(self, at: Timestamp) -> None:
        input_data = self._generate_input_data(update=True)
        input_data["data"]["data"]["id"] = self.id
        mutation_query = {"ok": None, "object": {"id": None}}
        query = Mutation(
            mutation=f"{self._schema.kind}Update",
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            at=at,
            tracker=f"mutation-{str(self._schema.kind).lower()}-update",
            variables=input_data["variables"],
        )


class NodeProperty:
    def __init__(self, data: Union[dict, str]):
        self.id = None
        self.display_label = None
        self.typename = None

        if isinstance(data, str):
            self.id = data
        elif isinstance(data, dict):
            self.id = data.get("id", None)
            self.display_label = data.get("display_label", None)
            self.typename = data.get("__typename", None)

    def _generate_input_data(self) -> Union[str, None]:
        return self.id


def generate_relationship_property(node: Union[InfrahubNode, InfrahubNodeSync], name: str, node_class):  # type: ignore
    """Return a property that stores values under a private non-public name."""
    internal_name = "_" + name.lower()
    external_name = name

    @property  # type: ignore
    def prop(self):  # type: ignore
        return getattr(self, internal_name)

    @prop.setter
    def prop(self, value):  # type: ignore
        if isinstance(value, RelatedNodeBase) or value is None:
            setattr(self, internal_name, value)
        else:
            schema = [rel for rel in self._schema.relationships if rel.name == external_name][0]
            setattr(
                self,
                internal_name,
                node_class(
                    name=external_name,
                    branch=node._branch,
                    client=node._client,
                    schema=schema,
                    data=value,
                ),
            )

    return prop



import glob
import importlib
import logging
import os
import sys
from typing import Optional

import typer
from rich.logging import RichHandler

app = typer.Typer()


INFRAHUB_CHECK_VARIABLE_TO_IMPORT = "INFRAHUB_CHECKS"


# pylint: disable=too-many-nested-blocks,too-many-branches


@app.callback()
def callback() -> None:
    """
    Execute user-defined checks.
    """


@app.command()
def run(
    branch: Optional[str] = None,
    path: Optional[str] = typer.Argument("."),
    rebase: bool = True,
    debug: bool = False,
    format_json: bool = False,
) -> None:
    """Locate and execute all checks under the defined path."""

    log_level = "DEBUG" if debug else "INFO"

    FORMAT = "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
    log = logging.getLogger("infrahub")

    python_files = glob.glob(f"{path}/**/*.py", recursive=True)

    passed_checks = []
    failed_checks = []

    nbr_checks_found = 0

    # search for python file
    for python_file in python_files:
        directory_name = os.path.dirname(python_file)
        filename = os.path.basename(python_file)
        module_name = os.path.splitext(filename)[0]

        if directory_name not in sys.path:
            sys.path.append(directory_name)

        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            continue

        if INFRAHUB_CHECK_VARIABLE_TO_IMPORT not in dir(module):
            continue

        for check_class in getattr(module, INFRAHUB_CHECK_VARIABLE_TO_IMPORT):
            nbr_checks_found += 1
            try:
                output = "stdout" if format_json else None
                check = check_class(output=output, root_directory=path, branch=branch, rebase=rebase)
                passed = check.run()

                if passed:
                    if not format_json:
                        log.info(f"{module_name}: [green]PASSED[/]", extra={"markup": True})
                    passed_checks.append(module_name)
                else:
                    failed_checks.append(module_name)
                    if not format_json:
                        log.error(f"{module_name}: [red]FAILED[/]", extra={"markup": True})

                        for log_message in check.logs:
                            log.error(f"  {log_message['message']}")

            except Exception as exc:  # pylint: disable=broad-except
                log.warning(f"{module_name}: An error occured during execution ({exc})")

    if nbr_checks_found == 0:
        if not format_json:
            log.warning("No check found")
        else:
            print('{"level": "WARNING", "message": "message", ""No check found"}')

    if failed_checks or not passed_checks:
        sys.exit(1)



import glob
import os
from pathlib import Path
from typing import List, Optional, Union

import pendulum
import yaml
from git import Repo
from pendulum.datetime import DateTime
from rich.console import Console
from rich.markup import escape

from infrahub_ctl.exceptions import FileNotValidError, QueryNotFoundError


def print_graphql_errors(console: Console, errors: List) -> None:
    if not isinstance(errors, list):
        console.print(f"[red]{escape(str(errors))}")

    for error in errors:
        if isinstance(error, dict) and "message" in error and "path" in error:
            console.print(f"[red]{escape(str(error['path']))} {escape(str(error['message']))}")
        else:
            console.print(f"[red]{escape(str(error))}")


def load_repository_config_file(repo_config_file: Path) -> dict:
    if not repo_config_file.is_file():
        raise FileNotFoundError(repo_config_file)

    try:
        yaml_data = repo_config_file.read_text()
        data = yaml.safe_load(yaml_data)
    except yaml.YAMLError as exc:
        raise FileNotValidError(name=str(repo_config_file)) from exc

    return data


def parse_cli_vars(variables: Optional[List[str]]) -> dict:
    if not variables:
        return {}

    return {var.split("=")[0]: var.split("=")[1] for var in variables if "=" in var}


def calculate_time_diff(value: str) -> Optional[str]:
    """Calculate the time in human format between a timedate in string format and now."""
    try:
        time_value = pendulum.parse(value)
    except pendulum.parsing.exceptions.ParserError:
        return None

    if not isinstance(time_value, DateTime):
        return None

    pendulum.set_locale("en")
    return time_value.diff_for_humans(other=pendulum.now(), absolute=True)


def find_graphql_query(name: str, directory: Union[str, Path] = ".") -> str:
    for query_file in glob.glob(f"{directory}/**/*.gql", recursive=True):
        filename = os.path.basename(query_file)
        query_name = os.path.splitext(filename)[0]

        if query_name != name:
            continue
        with open(query_file, "r", encoding="UTF-8") as file_data:
            query_string = file_data.read()

        return query_string

    raise QueryNotFoundError(name=name)


def render_action_rich(value: str) -> str:
    if value == "created":
        return f"[green]{value.upper()}[/green]"
    if value == "updated":
        return f"[magenta]{value.upper()}[/magenta]"
    if value == "deleted":
        return f"[red]{value.upper()}[/red]"

    return value.upper()


def get_branch(branch: Optional[str] = None, directory: Union[str, Path] = ".") -> str:
    """If branch isn't provide, return the name of the local Git branch."""
    if branch:
        return branch

    repo = Repo(directory)
    return str(repo.active_branch)


def get_fixtures_dir() -> Path:
    """Get the directory which stores fixtures that are common to multiple unit/integration tests."""
    here = os.path.abspath(os.path.dirname(__file__))
    fixtures_dir = os.path.join(here, "..", "tests", "fixtures")

    return Path(os.path.abspath(fixtures_dir))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from pydantic import BaseModel

from infrahub_sdk.exceptions import BranchNotFound
from infrahub_sdk.graphql import Mutation
from infrahub_sdk.queries import QUERY_ALL_BRANCHES, QUERY_BRANCH

if TYPE_CHECKING:
    from infrahub_sdk.client import InfrahubClient, InfrahubClientSync


class BranchData(BaseModel):
    id: str
    name: str
    description: Optional[str]
    is_data_only: bool
    is_default: bool
    origin_branch: Optional[str]
    branched_from: str


MUTATION_QUERY_DATA = {
    "ok": None,
    "object": {
        "id": None,
        "name": None,
        "description": None,
        "origin_branch": None,
        "branched_from": None,
        "is_default": None,
        "is_data_only": None,
    },
}


class InfraHubBranchManagerBase:
    @classmethod
    def generate_diff_data_url(
        cls,
        client: Union[InfrahubClient, InfrahubClientSync],
        branch_name: str,
        branch_only: bool = True,
        time_from: Optional[str] = None,
        time_to: Optional[str] = None,
    ) -> str:
        """Generate the URL for the diff_data function."""
        url = f"{client.address}/diff/data?branch={branch_name}"
        url += f"&branch_only={str(branch_only).lower()}"
        if time_from:
            url += f"&time_from={time_from}"
        if time_to:
            url += f"&time_to={time_to}"

        return url


class InfrahubBranchManager(InfraHubBranchManagerBase):
    def __init__(self, client: InfrahubClient):
        self.client = client

    async def create(
        self,
        branch_name: str,
        data_only: bool = False,
        description: str = "",
        background_execution: bool = False,
    ) -> BranchData:
        input_data = {
            "background_execution": background_execution,
            "data": {
                "name": branch_name,
                "description": description,
                "is_data_only": data_only,
            },
        }

        query = Mutation(mutation="BranchCreate", input_data=input_data, query=MUTATION_QUERY_DATA)
        response = await self.client.execute_graphql(query=query.render(), tracker="mutation-branch-create")

        return BranchData(**response["BranchCreate"]["object"])

    async def delete(self, branch_name: str) -> bool:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }
        query = Mutation(mutation="BranchDelete", input_data=input_data, query={"ok": None})
        response = await self.client.execute_graphql(query=query.render(), tracker="mutation-branch-delete")
        return response["BranchDelete"]["ok"]

    async def rebase(self, branch_name: str) -> BranchData:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }
        query = Mutation(mutation="BranchRebase", input_data=input_data, query=MUTATION_QUERY_DATA)
        response = await self.client.execute_graphql(query=query.render(), tracker="mutation-branch-rebase")
        return response["BranchRebase"]["ok"]

    async def validate(self, branch_name: str) -> BranchData:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }

        query_data = {
            "ok": None,
            "messages": None,
            "object": {
                "id": None,
                "name": None,
            },
        }

        query = Mutation(mutation="BranchValidate", input_data=input_data, query=query_data)
        response = await self.client.execute_graphql(query=query.render(), tracker="mutation-branch-validate")

        return response["BranchValidate"]["ok"]

    async def merge(self, branch_name: str) -> BranchData:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }
        query = Mutation(mutation="BranchMerge", input_data=input_data, query=MUTATION_QUERY_DATA)
        response = await self.client.execute_graphql(query=query.render(), tracker="mutation-branch-merge", timeout=120)

        return response["BranchMerge"]["ok"]

    async def all(self) -> Dict[str, BranchData]:
        data = await self.client.execute_graphql(query=QUERY_ALL_BRANCHES, tracker="query-branch-all")

        branches = {branch["name"]: BranchData(**branch) for branch in data["Branch"]}

        return branches

    async def get(self, branch_name: str) -> BranchData:
        data = await self.client.execute_graphql(
            query=QUERY_BRANCH,
            variables={"branch_name": branch_name},
            tracker="query-branch",
        )

        if not data["Branch"]:
            raise BranchNotFound(identifier=branch_name)
        return BranchData(**data["Branch"][0])

    async def diff_data(
        self,
        branch_name: str,
        branch_only: bool = True,
        time_from: Optional[str] = None,
        time_to: Optional[str] = None,
    ) -> Dict[Any, Any]:
        url = self.generate_diff_data_url(
            client=self.client,
            branch_name=branch_name,
            branch_only=branch_only,
            time_from=time_from,
            time_to=time_to,
        )
        response = await self.client._get(url=url, headers=self.client.headers)
        return response.json()


class InfrahubBranchManagerSync(InfraHubBranchManagerBase):
    def __init__(self, client: InfrahubClientSync):
        self.client = client

    def all(self) -> Dict[str, BranchData]:
        data = self.client.execute_graphql(query=QUERY_ALL_BRANCHES, tracker="query-branch-all")

        branches = {branch["name"]: BranchData(**branch) for branch in data["Branch"]}

        return branches

    def get(self, branch_name: str) -> BranchData:
        data = self.client.execute_graphql(
            query=QUERY_BRANCH,
            variables={"branch_name": branch_name},
            tracker="query-branch",
        )

        if not data["Branch"]:
            raise BranchNotFound(identifier=branch_name)
        return BranchData(**data["Branch"][0])

    def create(
        self,
        branch_name: str,
        data_only: bool = False,
        description: str = "",
        background_execution: bool = False,
    ) -> BranchData:
        input_data = {
            "background_execution": background_execution,
            "data": {
                "name": branch_name,
                "description": description,
                "is_data_only": data_only,
            },
        }

        query = Mutation(mutation="BranchCreate", input_data=input_data, query=MUTATION_QUERY_DATA)
        response = self.client.execute_graphql(query=query.render(), tracker="mutation-branch-create")

        return BranchData(**response["BranchCreate"]["object"])

    def delete(self, branch_name: str) -> bool:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }
        query = Mutation(mutation="BranchDelete", input_data=input_data, query={"ok": None})
        response = self.client.execute_graphql(query=query.render(), tracker="mutation-branch-delete")
        return response["BranchDelete"]["ok"]

    def diff_data(
        self,
        branch_name: str,
        branch_only: bool = True,
        time_from: Optional[str] = None,
        time_to: Optional[str] = None,
    ) -> Dict[Any, Any]:
        url = self.generate_diff_data_url(
            client=self.client,
            branch_name=branch_name,
            branch_only=branch_only,
            time_from=time_from,
            time_to=time_to,
        )
        response = self.client._get(url=url, headers=self.client.headers)
        return response.json()

    def merge(self, branch_name: str) -> BranchData:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }
        query = Mutation(mutation="BranchMerge", input_data=input_data, query=MUTATION_QUERY_DATA)
        response = self.client.execute_graphql(query=query.render(), tracker="mutation-branch-merge")

        return response["BranchMerge"]["ok"]

    def rebase(self, branch_name: str) -> BranchData:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }
        query = Mutation(mutation="BranchRebase", input_data=input_data, query=MUTATION_QUERY_DATA)
        response = self.client.execute_graphql(query=query.render(), tracker="mutation-branch-rebase")
        return response["BranchRebase"]["ok"]

    def validate(self, branch_name: str) -> BranchData:
        input_data = {
            "data": {
                "name": branch_name,
            }
        }

        query_data = {
            "ok": None,
            "messages": None,
            "object": {
                "id": None,
                "name": None,
            },
        }

        query = Mutation(mutation="BranchValidate", input_data=input_data, query=query_data)
        response = self.client.execute_graphql(query=query.render(), tracker="mutation-branch-validate")

        return response["BranchValidate"]["ok"]



from __future__ import annotations

import asyncio
import copy
import logging
from logging import Logger
from time import sleep
from typing import Any, Dict, List, MutableMapping, Optional, Union

import httpx

from infrahub_sdk.batch import InfrahubBatch
from infrahub_sdk.branch import (
    BranchData,
    InfrahubBranchManager,
    InfrahubBranchManagerSync,
)
from infrahub_sdk.config import Config
from infrahub_sdk.data import RepositoryData
from infrahub_sdk.exceptions import (
    AuthenticationError,
    GraphQLError,
    NodeNotFound,
    ServerNotReacheableError,
    ServerNotResponsiveError,
)
from infrahub_sdk.graphql import Query
from infrahub_sdk.node import InfrahubNode, InfrahubNodeSync
from infrahub_sdk.object_store import ObjectStore, ObjectStoreSync
from infrahub_sdk.queries import MUTATION_COMMIT_UPDATE, QUERY_ALL_REPOSITORIES
from infrahub_sdk.schema import InfrahubSchema, InfrahubSchemaSync, NodeSchema
from infrahub_sdk.store import NodeStore, NodeStoreSync
from infrahub_sdk.timestamp import Timestamp
from infrahub_sdk.types import AsyncRequester, HTTPMethod, SyncRequester
from infrahub_sdk.utils import is_valid_uuid

# pylint: disable=redefined-builtin


class BaseClient:
    """Base class for InfrahubClient and InfrahubClientSync"""

    def __init__(
        self,
        address: str = "",
        retry_on_failure: bool = False,
        retry_delay: int = 5,
        log: Optional[Logger] = None,
        default_branch: str = "main",
        insert_tracker: bool = False,
        pagination_size: int = 50,
        max_concurrent_execution: int = 5,
        config: Optional[Config] = None,
    ):
        self.client = None
        self.retry_on_failure = retry_on_failure
        self.retry_delay = retry_delay
        self.default_branch = default_branch
        self.log = log or logging.getLogger("infrahub_sdk")
        self.insert_tracker = insert_tracker
        self.pagination_size = pagination_size
        self.headers = {"content-type": "application/json"}
        self.access_token: str = ""
        self.refresh_token: str = ""
        if isinstance(config, Config):
            self.config = config
        elif isinstance(config, dict):
            self.config = Config(**config)
        else:
            self.config = Config()

        self.default_timeout = self.config.timeout
        self.config.address = address or self.config.address
        self.address = self.config.address

        if self.config.api_token:
            self.headers["X-INFRAHUB-KEY"] = self.config.api_token

        self.max_concurrent_execution = max_concurrent_execution

        self._initialize()

    def _initialize(self) -> None:
        """Sets the properties for each version of the client"""

    def _record(self, response: httpx.Response) -> None:
        if not self.config.custom_recorder:
            return
        self.config.custom_recorder.record(response)


class InfrahubClient(BaseClient):  # pylint: disable=too-many-public-methods
    """GraphQL Client to interact with Infrahub."""

    def _initialize(self) -> None:
        self.schema = InfrahubSchema(self)
        self.branch = InfrahubBranchManager(self)
        self.object_store = ObjectStore(self)
        self.store = NodeStore()
        self.concurrent_execution_limit = asyncio.Semaphore(self.max_concurrent_execution)
        self._request: AsyncRequester = self.config.requester or self._default_request_method

    @classmethod
    async def init(cls, *args: Any, **kwargs: Any) -> InfrahubClient:
        return cls(*args, **kwargs)

    async def create(
        self,
        kind: str,
        data: Optional[dict] = None,
        branch: Optional[str] = None,
        **kwargs: Any,
    ) -> InfrahubNode:
        branch = branch or self.default_branch
        schema = await self.schema.get(kind=kind, branch=branch)

        if not data and not kwargs:
            raise ValueError("Either data or a list of keywords but be provided")

        return InfrahubNode(client=self, schema=schema, branch=branch, data=data or kwargs)

    async def get(
        self,
        kind: str,
        at: Optional[Timestamp] = None,
        branch: Optional[str] = None,
        id: Optional[str] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        populate_store: bool = False,
        fragment: bool = False,
        **kwargs: Any,
    ) -> InfrahubNode:
        branch = branch or self.default_branch
        schema = await self.schema.get(kind=kind, branch=branch)

        filters: MutableMapping[str, Any] = {}

        if id:
            if not is_valid_uuid(id) and isinstance(schema, NodeSchema) and schema.default_filter:
                filters[schema.default_filter] = id
            else:
                filters["ids"] = [id]
        elif kwargs:
            filters = kwargs
        else:
            raise ValueError("At least one filter must be provided to get()")

        results = await self.filters(
            kind=kind,
            at=at,
            branch=branch,
            populate_store=populate_store,
            include=include,
            exclude=exclude,
            fragment=fragment,
            **filters,
        )  # type: ignore[arg-type]

        if len(results) == 0:
            raise NodeNotFound(branch_name=branch, node_type=kind, identifier=filters)
        if len(results) > 1:
            raise IndexError("More than 1 node returned")

        return results[0]

    async def all(
        self,
        kind: str,
        at: Optional[Timestamp] = None,
        branch: Optional[str] = None,
        populate_store: bool = False,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        fragment: bool = False,
    ) -> List[InfrahubNode]:
        """Retrieve all nodes of a given kind

        Args:
            kind (str): kind of the nodes to query
            at (Timestamp, optional): Time of the query. Defaults to Now.
            branch (str, optional): Name of the branch to query from. Defaults to default_branch.

        Returns:
            List[InfrahubNode]: List of Nodes
        """
        return await self.filters(
            kind=kind,
            at=at,
            branch=branch,
            populate_store=populate_store,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
            fragment=fragment,
        )

    async def filters(
        self,
        kind: str,
        at: Optional[Timestamp] = None,
        branch: Optional[str] = None,
        populate_store: bool = False,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        fragment: bool = False,
        **kwargs: Any,
    ) -> List[InfrahubNode]:
        schema = await self.schema.get(kind=kind)

        branch = branch or self.default_branch
        if at:
            at = Timestamp(at)

        node = InfrahubNode(client=self, schema=schema, branch=branch)
        filters = kwargs

        if filters:
            node.validate_filters(filters=filters)

        nodes = []
        # If Offset or Limit was provided we just query as it
        # If not, we'll query all nodes based on the size of the batch
        if offset or limit:
            query_data = await InfrahubNode(client=self, schema=schema, branch=branch).generate_query_data(
                offset=offset,
                limit=limit,
                filters=filters,
                include=include,
                exclude=exclude,
                fragment=fragment,
            )
            query = Query(query=query_data)
            response = await self.execute_graphql(
                query=query.render(),
                branch_name=branch,
                at=at,
                tracker=f"query-{str(schema.kind).lower()}-page1",
            )

            nodes = [
                await InfrahubNode.from_graphql(client=self, branch=branch, data=item)
                for item in response[schema.kind]["edges"]
            ]
        else:
            has_remaining_items = True
            page_number = 1
            while has_remaining_items:
                page_offset = (page_number - 1) * self.pagination_size

                query_data = await InfrahubNode(client=self, schema=schema, branch=branch).generate_query_data(
                    offset=page_offset,
                    limit=self.pagination_size,
                    filters=filters,
                    include=include,
                    exclude=exclude,
                    fragment=fragment,
                )
                query = Query(query=query_data)
                response = await self.execute_graphql(
                    query=query.render(),
                    branch_name=branch,
                    at=at,
                    tracker=f"query-{str(schema.kind).lower()}-page{page_number}",
                )

                nodes.extend(
                    [
                        await InfrahubNode.from_graphql(client=self, branch=branch, data=item)
                        for item in response[schema.kind]["edges"]
                    ]
                )

                remaining_items = response[schema.kind].get("count", 0) - (page_offset + self.pagination_size)
                if remaining_items < 0:
                    has_remaining_items = False

                page_number += 1

        if populate_store:
            for node in nodes:
                if node.id:
                    self.store.set(key=node.id, node=node)
        return nodes

    async def execute_graphql(  # pylint: disable=too-many-branches
        self,
        query: str,
        variables: Optional[dict] = None,
        branch_name: Optional[str] = None,
        at: Optional[Union[str, Timestamp]] = None,
        rebase: bool = False,
        timeout: Optional[int] = None,
        raise_for_error: bool = True,
        tracker: Optional[str] = None,
    ) -> Dict:
        """Execute a GraphQL query (or mutation).
        If retry_on_failure is True, the query will retry until the server becomes reacheable.

        Args:
            query (_type_): GraphQL Query to execute, can be a query or a mutation
            variables (dict, optional): Variables to pass along with the GraphQL query. Defaults to None.
            branch_name (str, optional): Name of the branch on which the query will be executed. Defaults to None.
            at (str, optional): Time when the query should be executed. Defaults to None.
            rebase (bool, optional): Flag to indicate if the branch should be rebased during the query. Defaults to False.
            timeout (int, optional): Timeout in second for the query. Defaults to None.
            raise_for_error (bool, optional): Flag to indicate that we need to raise an exception if the response has some errors. Defaults to True.

        Raises:
            GraphQLError: _description_

        Returns:
            _type_: _description_
        """

        url = f"{self.address}/graphql"
        if branch_name:
            url += f"/{branch_name}"

        payload: Dict[str, Union[str, dict]] = {"query": query}
        if variables:
            payload["variables"] = variables

        url_params = {}
        if at:
            at = Timestamp(at)
            url_params["at"] = at.to_string()

        if rebase:
            url_params["rebase"] = "true"
        if url_params:
            url += "?" + "&".join([f"{key}={value}" for key, value in url_params.items()])

        headers = copy.copy(self.headers or {})
        if self.insert_tracker and tracker:
            headers["X-Infrahub-Tracker"] = tracker

        # self.log.error(payload)

        retry = True
        while retry:
            retry = self.retry_on_failure
            try:
                resp = await self._post(url=url, payload=payload, headers=headers, timeout=timeout)

                if raise_for_error:
                    resp.raise_for_status()

                retry = False
            except ServerNotReacheableError:
                if retry:
                    self.log.warning(
                        f"Unable to connect to {self.address}, will retry in {self.retry_delay} seconds .."
                    )
                    await asyncio.sleep(delay=self.retry_delay)
                else:
                    self.log.error(f"Unable to connect to {self.address} .. ")
                    raise
            except httpx.HTTPStatusError as exc:
                if exc.response.status_code in [401, 403]:
                    response = exc.response.json()
                    errors = response.get("errors")
                    messages = [error.get("message") for error in errors]
                    raise AuthenticationError(" | ".join(messages)) from exc

        response = resp.json()

        if "errors" in response:
            raise GraphQLError(errors=response["errors"], query=query, variables=variables)

        return response["data"]

        # TODO add a special method to execute mutation that will check if the method returned OK

    async def _post(
        self,
        url: str,
        payload: dict,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
    ) -> httpx.Response:
        """Execute a HTTP POST with HTTPX.

        Raises:
            ServerNotReacheableError if we are not able to connect to the server
            ServerNotResponsiveError if the server didn't respond before the timeout expired
        """
        await self.login()
        headers = headers or {}
        base_headers = copy.copy(self.headers or {})
        headers.update(base_headers)
        return await self._request(
            url=url,
            method=HTTPMethod.POST,
            headers=headers,
            timeout=timeout or self.default_timeout,
            payload=payload,
        )

    async def _get(self, url: str, headers: Optional[dict] = None, timeout: Optional[int] = None) -> httpx.Response:
        """Execute a HTTP GET with HTTPX.

        Raises:
            ServerNotReacheableError if we are not able to connect to the server
            ServerNotResponsiveError if the server didnd't respond before the timeout expired
        """
        await self.login()
        headers = headers or {}
        base_headers = copy.copy(self.headers or {})
        headers.update(base_headers)
        return await self._request(
            url=url,
            method=HTTPMethod.GET,
            headers=headers,
            timeout=timeout or self.default_timeout,
        )

    async def _default_request_method(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        params: Dict[str, Any] = {}
        if payload:
            params["json"] = payload
        async with httpx.AsyncClient() as client:
            try:
                response = await client.request(
                    method=method.value,
                    url=url,
                    headers=headers,
                    timeout=timeout,
                    **params,
                )
            except httpx.NetworkError as exc:
                raise ServerNotReacheableError(address=self.address) from exc
            except httpx.ReadTimeout as exc:
                raise ServerNotResponsiveError(url=url, timeout=timeout) from exc

        self._record(response)
        return response

    async def login(self, refresh: bool = False) -> None:
        if not self.config.password_authentication:
            return

        if self.access_token and not refresh:
            return

        url = f"{self.address}/api/auth/login"
        response = await self._request(
            url=url,
            method=HTTPMethod.POST,
            payload={
                "username": self.config.username,
                "password": self.config.password,
            },
            headers={"content-type": "application/json"},
            timeout=self.default_timeout,
        )

        response.raise_for_status()
        self.access_token = response.json()["access_token"]
        self.refresh_token = response.json()["refresh_token"]
        self.headers["Authorization"] = f"Bearer {self.access_token}"

    async def query_gql_query(
        self,
        name: str,
        params: Optional[dict] = None,
        branch_name: Optional[str] = None,
        at: Optional[str] = None,
        rebase: bool = False,
        timeout: Optional[int] = None,
        raise_for_error: bool = True,
    ) -> Dict:
        url = f"{self.address}/api/query/{name}"
        url_params = copy.deepcopy(params or {})
        headers = copy.copy(self.headers or {})

        if branch_name:
            url_params["branch"] = branch_name
        if at:
            url_params["at"] = at
        if rebase:
            url_params["rebase"] = "true"

        if url_params:
            url += "?" + "&".join([f"{key}={value}" for key, value in url_params.items()])

        resp = await self._request(
            url=url,
            method=HTTPMethod.GET,
            headers=headers,
            timeout=timeout or self.default_timeout,
        )

        if raise_for_error:
            resp.raise_for_status()

        return resp.json()

    async def create_batch(self) -> InfrahubBatch:
        return InfrahubBatch(semaphore=self.concurrent_execution_limit)

    async def get_list_repositories(
        self, branches: Optional[Dict[str, BranchData]] = None
    ) -> Dict[str, RepositoryData]:
        if not branches:
            branches = await self.branch.all()  # type: ignore

        branch_names = sorted(branches.keys())  # type: ignore

        tasks = []
        for branch_name in branch_names:
            tasks.append(
                self.execute_graphql(
                    query=QUERY_ALL_REPOSITORIES,
                    branch_name=branch_name,
                    tracker="query-repository-all",
                )
            )
            # TODO need to rate limit how many requests we are sending at once to avoid doing a DOS on the API

        responses = await asyncio.gather(*tasks)

        repositories = {}

        for branch_name, response in zip(branch_names, responses):
            repos = response["CoreRepository"]["edges"]
            for repository in repos:
                repo_name = repository["node"]["name"]["value"]
                if repo_name not in repositories:
                    repositories[repo_name] = RepositoryData(
                        id=repository["node"]["id"],
                        name=repo_name,
                        location=repository["node"]["location"]["value"],
                        branches={},
                    )

                repositories[repo_name].branches[branch_name] = repository["node"]["commit"]["value"]

        return repositories

    async def repository_update_commit(self, branch_name: str, repository_id: str, commit: str) -> bool:
        variables = {"repository_id": str(repository_id), "commit": str(commit)}
        await self.execute_graphql(
            query=MUTATION_COMMIT_UPDATE,
            variables=variables,
            branch_name=branch_name,
            tracker="mutation-repository-update-commit",
        )

        return True


class InfrahubClientSync(BaseClient):  # pylint: disable=too-many-public-methods
    def _initialize(self) -> None:
        self.schema = InfrahubSchemaSync(self)
        self.branch = InfrahubBranchManagerSync(self)
        self.object_store = ObjectStoreSync(self)
        self.store = NodeStoreSync()
        self._request: SyncRequester = self.config.sync_requester or self._default_request_method

    @classmethod
    def init(cls, *args: Any, **kwargs: Any) -> InfrahubClientSync:
        return cls(*args, **kwargs)

    def create(
        self,
        kind: str,
        data: Optional[dict] = None,
        branch: Optional[str] = None,
        **kwargs: Any,
    ) -> InfrahubNodeSync:
        branch = branch or self.default_branch
        schema = self.schema.get(kind=kind, branch=branch)

        if not data and not kwargs:
            raise ValueError("Either data or a list of keywords but be provided")

        return InfrahubNodeSync(client=self, schema=schema, branch=branch, data=data or kwargs)

    def create_batch(self) -> InfrahubBatch:
        raise NotImplementedError("This method hasn't been implemented in the sync client yet.")

    def execute_graphql(  # pylint: disable=too-many-branches
        self,
        query: str,
        variables: Optional[dict] = None,
        branch_name: Optional[str] = None,
        at: Optional[Union[str, Timestamp]] = None,
        rebase: bool = False,
        timeout: Optional[int] = None,
        raise_for_error: bool = True,
        tracker: Optional[str] = None,
    ) -> Dict:
        """Execute a GraphQL query (or mutation).
        If retry_on_failure is True, the query will retry until the server becomes reacheable.

        Args:
            query (_type_): GraphQL Query to execute, can be a query or a mutation
            variables (dict, optional): Variables to pass along with the GraphQL query. Defaults to None.
            branch_name (str, optional): Name of the branch on which the query will be executed. Defaults to None.
            at (str, optional): Time when the query should be executed. Defaults to None.
            rebase (bool, optional): Flag to indicate if the branch should be rebased during the query. Defaults to False.
            timeout (int, optional): Timeout in second for the query. Defaults to None.
            raise_for_error (bool, optional): Flag to indicate that we need to raise an exception if the response has some errors. Defaults to True.

        Raises:
            GraphQLError: _description_

        Returns:
            _type_: _description_
        """

        url = f"{self.address}/graphql"
        if branch_name:
            url += f"/{branch_name}"

        payload: Dict[str, Union[str, dict]] = {"query": query}
        if variables:
            payload["variables"] = variables

        url_params = {}
        if at:
            at = Timestamp(at)
            url_params["at"] = at.to_string()

        if rebase:
            url_params["rebase"] = "true"
        if url_params:
            url += "?" + "&".join([f"{key}={value}" for key, value in url_params.items()])

        headers = copy.copy(self.headers or {})
        if self.insert_tracker and tracker:
            headers["X-Infrahub-Tracker"] = tracker

        retry = True
        while retry:
            retry = self.retry_on_failure
            try:
                resp = self._post(url=url, payload=payload, headers=headers, timeout=timeout)

                if raise_for_error:
                    resp.raise_for_status()

                retry = False
            except ServerNotReacheableError:
                if retry:
                    self.log.warning(
                        f"Unable to connect to {self.address}, will retry in {self.retry_delay} seconds .."
                    )
                    sleep(self.retry_delay)
                else:
                    self.log.error(f"Unable to connect to {self.address} .. ")
                    raise
            except httpx.HTTPStatusError as exc:
                if exc.response.status_code in [401, 403]:
                    response = exc.response.json()
                    errors = response.get("errors")
                    messages = [error.get("message") for error in errors]
                    raise AuthenticationError(" | ".join(messages)) from exc

        response = resp.json()

        if "errors" in response:
            raise GraphQLError(errors=response["errors"], query=query, variables=variables)

        return response["data"]

        # TODO add a special method to execute mutation that will check if the method returned OK

    def all(
        self,
        kind: str,
        at: Optional[Timestamp] = None,
        branch: Optional[str] = None,
        populate_store: bool = False,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        fragment: bool = False,
    ) -> List[InfrahubNodeSync]:
        """Retrieve all nodes of a given kind

        Args:
            kind (str): kind of the nodes to query
            at (Timestamp, optional): Time of the query. Defaults to Now.
            branch (str, optional): Name of the branch to query from. Defaults to default_branch.

        Returns:
            List[InfrahubNodeSync]: List of Nodes
        """

        return self.filters(
            kind=kind,
            at=at,
            branch=branch,
            populate_store=populate_store,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
            fragment=fragment,
        )

    def filters(
        self,
        kind: str,
        at: Optional[Timestamp] = None,
        branch: Optional[str] = None,
        populate_store: bool = False,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        fragment: bool = False,
        **kwargs: Any,
    ) -> List[InfrahubNodeSync]:
        schema = self.schema.get(kind=kind)

        branch = branch or self.default_branch
        if at:
            at = Timestamp(at)

        node = InfrahubNodeSync(client=self, schema=schema, branch=branch)
        filters = kwargs

        if filters:
            node.validate_filters(filters=filters)

        nodes = []
        # If Offset or Limit was provided we just query as it
        # If not, we'll query all nodes based on the size of the batch
        if offset or limit:
            query_data = InfrahubNodeSync(client=self, schema=schema, branch=branch).generate_query_data(
                offset=offset,
                limit=limit,
                filters=filters,
                include=include,
                exclude=exclude,
                fragment=fragment,
            )
            query = Query(query=query_data)
            response = self.execute_graphql(
                query=query.render(),
                branch_name=branch,
                at=at,
                tracker=f"query-{str(schema.kind).lower()}-page1",
            )

            nodes = [
                InfrahubNodeSync.from_graphql(client=self, branch=branch, data=item)
                for item in response[schema.kind]["edges"]
            ]

        else:
            has_remaining_items = True
            page_number = 1
            while has_remaining_items:
                page_offset = (page_number - 1) * self.pagination_size

                query_data = InfrahubNodeSync(client=self, schema=schema, branch=branch).generate_query_data(
                    offset=page_offset,
                    limit=self.pagination_size,
                    filters=filters,
                    include=include,
                    exclude=exclude,
                    fragment=fragment,
                )
                query = Query(query=query_data)
                response = self.execute_graphql(
                    query=query.render(),
                    branch_name=branch,
                    at=at,
                    tracker=f"query-{str(schema.kind).lower()}-page{page_number}",
                )

                nodes.extend(
                    [
                        InfrahubNodeSync.from_graphql(client=self, branch=branch, data=item)
                        for item in response[schema.kind]["edges"]
                    ]
                )

                remaining_items = response[schema.kind].get("count", 0) - (page_offset + self.pagination_size)
                if remaining_items < 0:
                    has_remaining_items = False

                page_number += 1

        if populate_store:
            for node in nodes:
                if node.id:
                    self.store.set(key=node.id, node=node)

        return nodes

    def get(
        self,
        kind: str,
        at: Optional[Timestamp] = None,
        branch: Optional[str] = None,
        id: Optional[str] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        populate_store: bool = False,
        fragment: bool = False,
        **kwargs: Any,
    ) -> InfrahubNodeSync:
        branch = branch or self.default_branch
        schema = self.schema.get(kind=kind, branch=branch)

        filters: MutableMapping[str, Any] = {}

        if id:
            if not is_valid_uuid(id) and isinstance(schema, NodeSchema) and schema.default_filter:
                filters[schema.default_filter] = id
            else:
                filters["ids"] = [id]
        elif kwargs:
            filters = kwargs
        else:
            raise ValueError("At least one filter must be provided to get()")

        results = self.filters(
            kind=kind,
            at=at,
            branch=branch,
            populate_store=populate_store,
            include=include,
            exclude=exclude,
            fragment=fragment,
            **filters,
        )  # type: ignore[arg-type]

        if len(results) == 0:
            raise NodeNotFound(branch_name=branch, node_type=kind, identifier=filters)
        if len(results) > 1:
            raise IndexError("More than 1 node returned")

        return results[0]

    def get_list_repositories(self, branches: Optional[Dict[str, BranchData]] = None) -> Dict[str, RepositoryData]:
        raise NotImplementedError(
            "This method is deprecated in the async client and won't be implemented in the sync client."
        )

    def query_gql_query(
        self,
        name: str,
        params: Optional[dict] = None,
        branch_name: Optional[str] = None,
        at: Optional[str] = None,
        rebase: bool = False,
        timeout: Optional[int] = None,
        raise_for_error: bool = True,
    ) -> Dict:
        raise NotImplementedError(
            "This method is deprecated in the async client and won't be implemented in the sync client."
        )

    def repository_update_commit(self, branch_name: str, repository_id: str, commit: str) -> bool:
        raise NotImplementedError(
            "This method is deprecated in the async client and won't be implemented in the sync client."
        )

    def _get(self, url: str, headers: Optional[dict] = None, timeout: Optional[int] = None) -> httpx.Response:
        """Execute a HTTP GET with HTTPX.

        Raises:
            ServerNotReacheableError if we are not able to connect to the server
            ServerNotResponsiveError if the server didnd't respond before the timeout expired
        """
        self.login()
        headers = headers or {}
        base_headers = copy.copy(self.headers or {})
        headers.update(base_headers)
        return self._request(
            url=url,
            method=HTTPMethod.GET,
            headers=headers,
            timeout=timeout or self.default_timeout,
        )

    def _post(
        self,
        url: str,
        payload: dict,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
    ) -> httpx.Response:
        """Execute a HTTP POST with HTTPX.

        Raises:
            ServerNotReacheableError if we are not able to connect to the server
            ServerNotResponsiveError if the server didnd't respond before the timeout expired
        """
        self.login()
        headers = headers or {}
        base_headers = copy.copy(self.headers or {})
        headers.update(base_headers)

        return self._request(
            url=url,
            method=HTTPMethod.POST,
            payload=payload,
            headers=headers,
            timeout=timeout or self.default_timeout,
        )

    def _default_request_method(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        params: Dict[str, Any] = {}
        if payload:
            params["json"] = payload
        with httpx.Client() as client:
            try:
                response = client.request(
                    method=method.value,
                    url=url,
                    headers=headers,
                    timeout=timeout,
                    **params,
                )
            except httpx.NetworkError as exc:
                raise ServerNotReacheableError(address=self.address) from exc
            except httpx.ReadTimeout as exc:
                raise ServerNotResponsiveError(url=url, timeout=timeout) from exc

        self._record(response)
        return response

    def login(self, refresh: bool = False) -> None:
        if not self.config.password_authentication:
            return

        if self.access_token and not refresh:
            return

        url = f"{self.address}/api/auth/login"
        response = self._request(
            url=url,
            method=HTTPMethod.POST,
            payload={
                "username": self.config.username,
                "password": self.config.password,
            },
            headers={"content-type": "application/json"},
            timeout=self.default_timeout,
        )

        response.raise_for_status()
        self.access_token = response.json()["access_token"]
        self.refresh_token = response.json()["refresh_token"]
        self.headers["Authorization"] = f"Bearer {self.access_token}"



from __future__ import annotations

from collections import defaultdict
from enum import Enum
from pathlib import Path  # noqa: TCH003
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Union,
)

from pydantic import BaseModel, Field

from infrahub_sdk.exceptions import SchemaNotFound, ValidationError

if TYPE_CHECKING:
    from infrahub_sdk.client import InfrahubClient, InfrahubClientSync

# pylint: disable=redefined-builtin


# ---------------------------------------------------------------------------------
# Repository Configuration file
# ---------------------------------------------------------------------------------
class InfrahubRepositoryRFileConfig(BaseModel):
    name: str
    query: str
    repository: str
    template_path: Path


class InfrahubRepositoryConfig(BaseModel):
    schemas: List[Path] = Field(default_factory=list)
    rfiles: Optional[List[InfrahubRepositoryRFileConfig]]


# ---------------------------------------------------------------------------------
# Main Infrahub Schema File
# ---------------------------------------------------------------------------------
class FilterSchema(BaseModel):
    name: str
    kind: str
    description: Optional[str]


class RelationshipCardinality(str, Enum):
    ONE = "one"
    MANY = "many"


class BranchSupportType(str, Enum):
    AWARE = "aware"
    AGNOSTIC = "agnostic"
    LOCAL = "local"


class RelationshipKind(str, Enum):
    GENERIC = "Generic"
    ATTRIBUTE = "Attribute"
    COMPONENT = "Component"
    PARENT = "Parent"
    GROUP = "Group"


class AttributeSchema(BaseModel):
    name: str
    kind: str
    label: Optional[str]
    description: Optional[str]
    default_value: Optional[Any]
    inherited: bool = False
    unique: bool = False
    branch: Optional[BranchSupportType]
    optional: bool = False


class RelationshipSchema(BaseModel):
    name: str
    peer: str
    kind: RelationshipKind = RelationshipKind.GENERIC
    label: Optional[str]
    description: Optional[str]
    identifier: Optional[str]
    inherited: bool = False
    cardinality: str = "many"
    branch: Optional[BranchSupportType]
    optional: bool = True
    filters: List[FilterSchema] = Field(default_factory=list)


class BaseNodeSchema(BaseModel):
    name: str
    label: Optional[str]
    namespace: str
    description: Optional[str]
    attributes: List[AttributeSchema] = Field(default_factory=list)
    relationships: List[RelationshipSchema] = Field(default_factory=list)
    filters: List[FilterSchema] = Field(default_factory=list)

    @property
    def kind(self) -> str:
        return self.namespace + self.name

    def get_field(self, name: str, raise_on_error: bool = True) -> Union[AttributeSchema, RelationshipSchema, None]:
        if attribute_field := self.get_attribute(name, raise_on_error=False):
            return attribute_field

        if relationship_field := self.get_relationship(name, raise_on_error=False):
            return relationship_field

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the field {name}")

    def get_attribute(self, name: str, raise_on_error: bool = True) -> Union[AttributeSchema, None]:
        for item in self.attributes:
            if item.name == name:
                return item

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the attribute {name}")

    def get_relationship(self, name: str, raise_on_error: bool = True) -> Union[RelationshipSchema, None]:
        for item in self.relationships:
            if item.name == name:
                return item

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the relationship {name}")

    def get_relationship_by_identifier(self, id: str, raise_on_error: bool = True) -> Union[RelationshipSchema, None]:
        for item in self.relationships:
            if item.identifier == id:
                return item

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the relationship {id}")

    @property
    def attribute_names(self) -> List[str]:
        return [item.name for item in self.attributes]

    @property
    def relationship_names(self) -> List[str]:
        return [item.name for item in self.relationships]

    @property
    def mandatory_input_names(self) -> List[str]:
        return self.mandatory_attribute_names + self.mandatory_relationship_names

    @property
    def mandatory_attribute_names(self) -> List[str]:
        return [item.name for item in self.attributes if not item.optional and item.default_value is None]

    @property
    def mandatory_relationship_names(self) -> List[str]:
        return [item.name for item in self.relationships if not item.optional]

    @property
    def local_attributes(self) -> List[AttributeSchema]:
        return [item for item in self.attributes if not item.inherited]

    @property
    def local_relationships(self) -> List[RelationshipSchema]:
        return [item for item in self.relationships if not item.inherited]

    @property
    def unique_attributes(self) -> List[AttributeSchema]:
        return [item for item in self.attributes if item.unique]


class GenericSchema(BaseNodeSchema):
    """A Generic can be either an Interface or a Union depending if there are some Attributes or Relationships defined."""

    used_by: List[str] = Field(default_factory=list)


class NodeSchema(BaseNodeSchema):
    inherit_from: List[str] = Field(default_factory=list)
    groups: List[str] = Field(default_factory=list)
    branch: Optional[BranchSupportType]
    default_filter: Optional[str]


class NodeExtensionSchema(BaseModel):
    name: Optional[str]
    kind: str
    description: Optional[str]
    label: Optional[str]
    inherit_from: List[str] = Field(default_factory=list)
    groups: List[str] = Field(default_factory=list)
    branch: Optional[BranchSupportType]
    default_filter: Optional[str]
    attributes: List[AttributeSchema] = Field(default_factory=list)
    relationships: List[RelationshipSchema] = Field(default_factory=list)


class GroupSchema(BaseModel):
    name: str
    kind: str
    description: Optional[str]


class SchemaRoot(BaseModel):
    version: str
    generics: List[GenericSchema] = Field(default_factory=list)
    nodes: List[NodeSchema] = Field(default_factory=list)
    groups: List[GroupSchema] = Field(default_factory=list)
    # node_extensions: List[NodeExtensionSchema] = Field(default_factory=list)


class InfrahubSchemaBase:
    def validate(self, data: dict[str, Any]) -> None:
        SchemaRoot(**data)

    def validate_data_against_schema(self, schema: Union[NodeSchema, GenericSchema], data: dict) -> None:
        for key in data.keys():
            if key not in schema.relationship_names + schema.attribute_names:
                identifier = f"{schema.kind}"
                raise ValidationError(
                    identifier=identifier,
                    message=f"{key} is not a valid value for {identifier}",
                )

    def generate_payload_create(
        self,
        schema: Union[NodeSchema, GenericSchema],
        data: dict,
        source: Optional[str] = None,
        owner: Optional[str] = None,
        is_protected: Optional[bool] = None,
        is_visible: Optional[bool] = None,
    ) -> Dict[str, Any]:
        obj_data: Dict[str, Any] = {}
        item_metadata: Dict[str, Any] = {}
        if source:
            item_metadata["source"] = str(source)
        if owner:
            item_metadata["owner"] = str(owner)
        if is_protected is not None:
            item_metadata["is_protected"] = is_protected
        if is_visible is not None:
            item_metadata["is_visible"] = is_visible

        for key, value in data.items():
            obj_data[key] = {}
            if key in schema.attribute_names:
                obj_data[key] = {"value": value}
                obj_data[key].update(item_metadata)
            elif key in schema.relationship_names:
                rel = schema.get_relationship(name=key)
                if rel:
                    if rel.cardinality == "one":
                        obj_data[key] = {"id": str(value)}
                        obj_data[key].update(item_metadata)
                    elif rel.cardinality == "many":
                        obj_data[key] = [{"id": str(item)} for item in value]
                        for item in obj_data[key]:
                            item.update(item_metadata)

        return obj_data


class InfrahubSchema(InfrahubSchemaBase):
    """
    client.schema.get(branch="name", kind="xxx")
    client.schema.all(branch="xxx")
    client.schema.validate()
    client.schema.add()
    client.schema.node.add()
    """

    def __init__(self, client: InfrahubClient):
        self.client = client
        self.cache: dict = defaultdict(lambda: dict)

    async def get(
        self, kind: str, branch: Optional[str] = None, refresh: bool = False
    ) -> Union[NodeSchema, GenericSchema]:
        branch = branch or self.client.default_branch

        if refresh:
            self.cache[branch] = await self.fetch(branch=branch)

        if branch in self.cache and kind in self.cache[branch]:
            return self.cache[branch][kind]

        # Fetching the latest schema from the server if we didn't fetch it earlier
        #   because we coulnd't find the object on the local cache
        if not refresh:
            self.cache[branch] = await self.fetch(branch=branch)

        if branch in self.cache and kind in self.cache[branch]:
            return self.cache[branch][kind]

        raise SchemaNotFound(identifier=kind)

    async def all(
        self, branch: Optional[str] = None, refresh: bool = False
    ) -> MutableMapping[str, Union[NodeSchema, GenericSchema]]:
        """Retrieve the entire schema for a given branch.

        if present in cache, the schema will be served from the cache, unless refresh is set to True
        if the schema is not present in the cache, it will be fetched automatically from the server

        Args:
            branch (str, optional): Name of the branch to query. Defaults to default_branch.
            refresh (bool, optional): Force a refresh of the schema. Defaults to False.

        Returns:
            Dict[str, NodeSchema]: Dictionnary of all schema organized by kind
        """
        branch = branch or self.client.default_branch
        if refresh or branch not in self.cache:
            self.cache[branch] = await self.fetch(branch=branch)

        return self.cache[branch]

    async def load(self, schemas: List[dict], branch: Optional[str] = None) -> Tuple[bool, Optional[dict]]:
        branch = branch or self.client.default_branch
        url = f"{self.client.address}/api/schema/load?branch={branch}"
        response = await self.client._post(url=url, timeout=60, payload={"schemas": schemas})

        if response.status_code == 202:
            return True, None

        if response.status_code == 422:
            return False, response.json()

        response.raise_for_status()
        return False, None

    async def fetch(self, branch: str) -> MutableMapping[str, Union[NodeSchema, GenericSchema]]:
        """Fetch the schema from the server for a given branch.

        Args:
            branch (str): Name of the branch to fetch the schema for.

        Returns:
            Dict[str, NodeSchema]: Dictionnary of all schema organized by kind
        """
        url = f"{self.client.address}/api/schema/?branch={branch}"
        response = await self.client._get(url=url)
        response.raise_for_status()

        data: MutableMapping[str, Any] = response.json()

        nodes: MutableMapping[str, Union[NodeSchema, GenericSchema]] = {}
        for node_schema in data.get("nodes", []):
            node = NodeSchema(**node_schema)
            nodes[node.kind] = node

        for generic_schema in data.get("generics", []):
            generic = GenericSchema(**generic_schema)
            nodes[generic.kind] = generic

        return nodes


class InfrahubSchemaSync(InfrahubSchemaBase):
    def __init__(self, client: InfrahubClientSync):
        self.client = client
        self.cache: dict = defaultdict(lambda: dict)

    def all(
        self, branch: Optional[str] = None, refresh: bool = False
    ) -> MutableMapping[str, Union[NodeSchema, GenericSchema]]:
        """Retrieve the entire schema for a given branch.

        if present in cache, the schema will be served from the cache, unless refresh is set to True
        if the schema is not present in the cache, it will be fetched automatically from the server

        Args:
            branch (str, optional): Name of the branch to query. Defaults to default_branch.
            refresh (bool, optional): Force a refresh of the schema. Defaults to False.

        Returns:
            Dict[str, NodeSchema]: Dictionnary of all schema organized by kind
        """
        branch = branch or self.client.default_branch
        if refresh or branch not in self.cache:
            self.cache[branch] = self.fetch(branch=branch)

        return self.cache[branch]

    def get(self, kind: str, branch: Optional[str] = None, refresh: bool = False) -> Union[NodeSchema, GenericSchema]:
        branch = branch or self.client.default_branch

        if refresh:
            self.cache[branch] = self.fetch(branch=branch)

        if branch in self.cache and kind in self.cache[branch]:
            return self.cache[branch][kind]

        # Fetching the latest schema from the server if we didn't fetch it earlier
        #   because we coulnd't find the object on the local cache
        if not refresh:
            self.cache[branch] = self.fetch(branch=branch)

        if branch in self.cache and kind in self.cache[branch]:
            return self.cache[branch][kind]

        raise SchemaNotFound(identifier=kind)

    def fetch(self, branch: str) -> MutableMapping[str, Union[NodeSchema, GenericSchema]]:
        """Fetch the schema from the server for a given branch.

        Args:
            branch (str): Name of the branch to fetch the schema for.

        Returns:
            Dict[str, NodeSchema]: Dictionnary of all schema organized by kind
        """
        url = f"{self.client.address}/api/schema/?branch={branch}"
        response = self.client._get(url=url)
        response.raise_for_status()

        data: MutableMapping[str, Any] = response.json()

        nodes: MutableMapping[str, Union[NodeSchema, GenericSchema]] = {}
        for node_schema in data.get("nodes", []):
            node = NodeSchema(**node_schema)
            nodes[node.kind] = node

        for generic_schema in data.get("generics", []):
            generic = GenericSchema(**generic_schema)
            nodes[generic.kind] = generic

        return nodes

    def load(self, schemas: List[dict], branch: Optional[str] = None) -> Tuple[bool, Optional[dict]]:
        branch = branch or self.client.default_branch
        url = f"{self.client.address}/api/schema/load?branch={branch}"
        response = self.client._post(url=url, timeout=60, payload={"schemas": schemas})

        if response.status_code == 202:
            return True, None

        if response.status_code == 422:
            return False, response.json()

        response.raise_for_status()
        return False, None



import sys
from asyncio import run as aiorun
from pathlib import Path
from typing import List, Optional

import typer
import yaml
from pydantic import ValidationError
from rich.console import Console
from ujson import JSONDecodeError

import infrahub_ctl.config as config
from infrahub_ctl.client import initialize_client, initialize_client_sync
from infrahub_ctl.exceptions import QueryNotFoundError
from infrahub_ctl.utils import find_graphql_query, get_branch, parse_cli_vars
from infrahub_sdk.exceptions import GraphQLError

app = typer.Typer()


@app.callback()
def callback() -> None:
    """
    Helper to validate the format of various files.
    """


async def _schema(schema: Path) -> None:
    console = Console()

    try:
        schema_data = yaml.safe_load(schema.read_text())
    except JSONDecodeError as exc:
        console.print("[red]Invalid JSON file")
        raise typer.Exit(2) from exc

    client = await initialize_client()

    try:
        client.schema.validate(schema_data)
    except ValidationError as exc:
        console.print(f"[red]Schema not valid, found '{len(exc.errors())}' error(s)")
        for error in exc.errors():
            loc_str = [str(item) for item in error["loc"]]
            console.print(f"  '{'/'.join(loc_str)}' | {error['msg']} ({error['type']})")
        raise typer.Exit(2)

    console.print("[green]Schema is valid !!")


@app.command(name="schema")
def validate_schema(
    schema: Path,
    config_file: Path = typer.Option(config.DEFAULT_CONFIG_FILE, envvar=config.ENVVAR_CONFIG_FILE),
) -> None:
    """Validate the format of a schema file either in JSON or YAML"""

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_schema(schema=schema))


@app.command(name="graphql-query")
def validate_graphql(
    query: str,
    variables: Optional[List[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
    debug: bool = typer.Option(False, help="Display more troubleshooting information."),
    branch: str = typer.Option(None, help="Branch on which to validate the GraphQL Query."),
    config_file: Path = typer.Option(config.DEFAULT_CONFIG_FILE, envvar=config.ENVVAR_CONFIG_FILE),
) -> None:
    """Validate the format of a GraphQL Query stored locally by executing it on a remote GraphQL endpoint"""

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    console = Console()

    branch = get_branch(branch)

    try:
        query_str = find_graphql_query(query)
    except QueryNotFoundError:
        console.print(f"[red]Unable to find the GraphQL Query : {query}")
        sys.exit(1)

    console.print(f"[purple]Query '{query}' will be validated on branch '{branch}'.")

    variables_dict = parse_cli_vars(variables)

    client = initialize_client_sync()
    try:
        response = client.execute_graphql(
            query=query_str,
            branch_name=branch,
            variables=variables_dict,
            raise_for_error=False,
        )
    except GraphQLError as exc:
        console.print(f"[red]{len(exc.errors)} error(s) occured while executing the query")
        for error in exc.errors:
            if isinstance(error, dict) and "message" in error and "locations" in error:
                console.print(f"[yellow] - Message: {error['message']}")
                console.print(f"[yellow]   Location: {error['locations']}")
            elif isinstance(error, str) and "Branch:" in error:
                console.print(f"[yellow] - {error}")
                console.print("[yellow]   you can specify a different branch with --branch")
        sys.exit(1)

    console.print("[green] Query executed successfully.")

    if debug:
        console.print("-" * 40)
        console.print(f"Response for GraphQL Query {query}")
        console.print(response)
        console.print("-" * 40)



import glob
import hashlib
import os
from itertools import groupby
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import httpx
import yaml
from pydantic import BaseModel


class YamlFile(BaseModel):
    identifier: str
    location: Path
    content: Optional[dict] = None
    valid: bool = True
    error_message: Optional[str] = None

    def load_content(self) -> None:
        try:
            self.content = yaml.safe_load(self.location.read_text())
        except yaml.YAMLError:
            self.error_message = "Invalid YAML/JSON file"
            self.valid = False


def base36encode(number: int) -> str:
    if not isinstance(number, (int)):
        raise TypeError("number must be an integer")
    is_negative = number < 0
    number = abs(number)

    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    base36 = ""

    while number:
        number, i = divmod(number, 36)
        base36 = alphabet[i] + base36
    if is_negative:
        base36 = "-" + base36

    return base36 or alphabet[0]


def base36decode(data: str) -> int:
    return int(data, 36)


def base16decode(data: str) -> int:
    return int(data, 16)


def base16encode(number: int) -> str:
    if not isinstance(number, (int)):
        raise TypeError("number must be an integer")
    is_negative = number < 0
    number = abs(number)

    alphabet = "0123456789abcdef"
    base16 = ""

    while number:
        number, i = divmod(number, 16)
        base16 = alphabet[i] + base16
    if is_negative:
        base16 = "-" + base16

    return base16 or alphabet[0]


def get_fixtures_dir() -> Path:
    """Get the directory which stores fixtures that are common to multiple unit/integration tests."""
    here = os.path.abspath(os.path.dirname(__file__))
    fixtures_dir = os.path.join(here, "..", "tests", "fixtures")

    return Path(os.path.abspath(fixtures_dir))


def is_valid_uuid(value: Any) -> bool:
    """Check if the input is a valid UUID."""
    try:
        UUID(str(value))
        return True
    except ValueError:
        return False


def generate_uuid() -> str:
    return str(uuid4())


def duplicates(input_list: list) -> list:
    """Identify and return all the duplicates in a list."""

    dups = []

    for x, y in groupby(sorted(input_list)):
        #  list(y) returns all the occurences of item x
        if len(list(y)) > 1:
            dups.append(x)

    return dups


def intersection(list1: List[Any], list2: List[Any]) -> list:
    """Calculate the intersection between 2 lists."""
    return list(set(list1) & set(list2))


def compare_lists(list1: List[Any], list2: List[Any]) -> Tuple[List[Any], List[Any], List[Any]]:
    """Compare 2 lists and return :
    - the intersection of both
    - the item present only in list1
    - the item present only in list2
    """

    in_both = intersection(list1=list1, list2=list2)
    in_list_1 = list(set(list1) - set(in_both))
    in_list_2 = list(set(list2) - set(in_both))

    return sorted(in_both), sorted(in_list_1), sorted(in_list_2)


def deep_merge_dict(dicta: dict, dictb: dict, path: Optional[List] = None) -> dict:
    """Deep Merge dictionnary B into Dictionnary A.
    Code is inspired by https://stackoverflow.com/a/7205107
    """
    if path is None:
        path = []
    for key in dictb:
        if key in dicta:
            if isinstance(dicta[key], dict) and isinstance(dictb[key], dict):
                deep_merge_dict(dicta[key], dictb[key], path + [str(key)])
            elif dicta[key] == dictb[key]:
                pass
            else:
                raise ValueError("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            dicta[key] = dictb[key]
    return dicta


def str_to_bool(value: str) -> bool:
    """Convert a String to a Boolean"""

    if isinstance(value, bool):
        return value

    if isinstance(value, int) and value in [0, 1]:
        return bool(value)

    if not isinstance(value, str):
        raise TypeError(f"{value} must be a string")

    MAP = {
        "y": True,
        "yes": True,
        "t": True,
        "true": True,
        "on": True,
        "1": True,
        "n": False,
        "no": False,
        "f": False,
        "false": False,
        "off": False,
        "0": False,
    }
    try:
        return MAP[value.lower()]
    except KeyError as exc:
        raise ValueError(f"{value} can not be converted into a boolean") from exc


def get_flat_value(obj: Any, key: str, separator: str = "__") -> Any:
    """Query recursively an value defined in a flat notation (string), on a hierarchy of objects

    Examples:
        name__value
        module.object.value
    """
    if separator not in key:
        return getattr(obj, key)

    first_part, remaining_part = key.split(separator, maxsplit=1)
    sub_obj = getattr(obj, first_part)
    if not sub_obj:
        return None
    return get_flat_value(obj=sub_obj, key=remaining_part, separator=separator)


def generate_request_filename(request: httpx.Request) -> str:
    """Return a filename for a request sent to the Infrahub API

    This function is used when recording and playing back requests, as Infrahub is using a GraphQL
    API it's not possible to rely on the URL endpoint alone to separate one request from another,
    for this reason a hash of the payload is included in a filename.
    """
    formatted = (
        str(request.url).replace(":", "_").replace("//", "").replace("/", "__").replace("?", "_q_").replace("&", "_a_")
    )
    filename = f"{request.method}_{formatted}"
    if request.content:
        content_hash = hashlib.sha224(request.content)
        filename += f"_{content_hash.hexdigest()}"

    return filename.lower()


def is_valid_url(url: str) -> bool:
    try:
        parsed = httpx.URL(url)
        return all([parsed.scheme, parsed.netloc])
    except TypeError:
        return False


def find_files(
    extension: Union[str, List[str]],
    directory: Union[str, Path] = ".",
    recursive: bool = True,
) -> List[Path]:
    files = []

    if isinstance(extension, str):
        extension = [extension]

    for ext in extension:
        files.extend(glob.glob(f"{directory}/**/*.{ext}", recursive=recursive))
        files.extend(glob.glob(f"{directory}/**/.*.{ext}", recursive=recursive))

    return [Path(item) for item in files]



import importlib
import linecache
import logging
import os
import sys
from asyncio import run as aiorun
from pathlib import Path
from typing import List, Optional, Tuple

import jinja2
import typer
from pydantic import ValidationError
from rich.console import Console
from rich.logging import RichHandler
from rich.syntax import Syntax
from rich.traceback import Frame, Traceback

# pylint: disable=import-outside-toplevel
import infrahub_ctl.config as config
from infrahub_ctl.branch import app as branch_app
from infrahub_ctl.check import app as check_app
from infrahub_ctl.client import initialize_client, initialize_client_sync
from infrahub_ctl.exceptions import FileNotValidError, QueryNotFoundError
from infrahub_ctl.schema import app as schema
from infrahub_ctl.utils import (
    find_graphql_query,
    get_branch,
    load_repository_config_file,
    parse_cli_vars,
)
from infrahub_ctl.validate import app as validate_app
from infrahub_sdk.exceptions import GraphQLError
from infrahub_sdk.schema import InfrahubRepositoryConfig

app = typer.Typer(pretty_exceptions_show_locals=False)

app.add_typer(branch_app, name="branch")
app.add_typer(check_app, name="check")
app.add_typer(schema, name="schema")
app.add_typer(validate_app, name="validate")


async def _run(
    script: Path,
    method: str,
    log: logging.Logger,
    branch: str,
    concurrent: int,
    timeout: int,
) -> None:
    directory_name = os.path.dirname(script)
    filename = os.path.basename(script)
    module_name = os.path.splitext(filename)[0]

    if directory_name not in sys.path:
        sys.path.append(directory_name)

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        raise typer.Abort(f"Unable to Load the Python script at {script}") from exc

    if not hasattr(module, method):
        raise typer.Abort(f"Unable to Load the method {method} in the Python script at {script}")

    client = await initialize_client(timeout=timeout, max_concurrent_execution=concurrent)

    func = getattr(module, method)
    await func(client=client, log=log, branch=branch)


def identify_faulty_jinja_code(traceback: Traceback, nbr_context_lines: int = 3) -> List[Tuple[Frame, Syntax]]:
    response = []

    # The Traceback from rich is very helpfull to parse the entire stack trace
    # to will generate a Frame object for each exception in the trace

    # Extract only the Jinja related exceptioin from the stack
    frames = [frame for frame in traceback.trace.stacks[0].frames if frame.filename.endswith(".j2")]

    for frame in frames:
        code = "".join(linecache.getlines(frame.filename))
        lexer_name = Traceback._guess_lexer(frame.filename, code)
        syntax = Syntax(
            code,
            lexer_name,
            line_numbers=True,
            line_range=(
                frame.lineno - nbr_context_lines,
                frame.lineno + nbr_context_lines,
            ),
            highlight_lines={frame.lineno},
            code_width=88,
            theme=traceback.theme,
            dedent=False,
        )
        response.append((frame, syntax))

    return response


@app.command()
def render(  # pylint: disable=too-many-branches,too-many-statements
    rfile: str,
    variables: Optional[List[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
    branch: str = typer.Option(None, help="Branch on which to rendre the RFile."),
    debug: bool = False,
    config_file: str = typer.Option(config.DEFAULT_CONFIG_FILE, envvar=config.ENVVAR_CONFIG_FILE),
) -> None:
    """Render a local Jinja Template (RFile) for debugging purpose."""

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    branch = get_branch(branch)

    console = Console()

    # ------------------------------------------------------------------
    # Read the configuration file
    # ------------------------------------------------------------------
    try:
        config_file_data = load_repository_config_file(repo_config_file=Path(config.INFRAHUB_REPO_CONFIG_FILE))
    except FileNotFoundError as exc:
        console.print(f"[red]{exc}")
        raise typer.Exit(1) from exc
    except FileNotValidError as exc:
        console.print(f"[red]{exc}")
        raise typer.Exit(1) from exc

    try:
        data = InfrahubRepositoryConfig(**config_file_data)
    except ValidationError as exc:
        console.print(f"[red]Repository config file not valid, found {len(exc.errors())} error(s)")
        for error in exc.errors():
            loc_str = [str(item) for item in error["loc"]]
            console.print(f"  {'/'.join(loc_str)} | {error['msg']} ({error['type']})")
        raise typer.Exit(1) from exc

    # ------------------------------------------------------------------
    # Find the GraphQL Query and Retrive its data
    # ------------------------------------------------------------------
    filtered_rfile = [entry for entry in data.rfiles if entry.name == rfile]
    if not filtered_rfile:
        console.print(f"[red]Unable to find {rfile} in {config.INFRAHUB_REPO_CONFIG_FILE}")
        raise typer.Exit(1)

    rfile_data = filtered_rfile[0]

    try:
        query_str = find_graphql_query(rfile_data.query)
    except QueryNotFoundError as exc:
        console.print(f"[red]Unable to find query : {exc}")
        raise typer.Exit(1) from exc

    variables_dict = parse_cli_vars(variables)

    client = initialize_client_sync()
    try:
        response = client.execute_graphql(
            query=query_str,
            branch_name=branch,
            variables=variables_dict,
            raise_for_error=False,
        )
    except GraphQLError as exc:
        console.print(f"[red]{len(exc.errors)} error(s) occured while executing the query")
        for error in exc.errors:
            if isinstance(error, dict) and "message" in error and "locations" in error:
                console.print(f"[yellow] - Message: {error['message']}")  # type: ignore[typeddict-item]
                console.print(f"[yellow]   Location: {error['locations']}")  # type: ignore[typeddict-item]
            elif isinstance(error, str) and "Branch:" in error:
                console.print(f"[yellow] - {error}")
                console.print("[yellow]   you can specify a different branch with --branch")
        raise typer.Abort()

    if debug:
        console.print("-" * 40)
        console.print(f"Response for GraphQL Query {rfile_data.query}")
        console.print(response)
        console.print("-" * 40)

    # ------------------------------------------------------------------
    # Finally, render the template
    # ------------------------------------------------------------------
    template_path = rfile_data.template_path
    if not template_path.is_file():
        console.print(f"[red]Unable to locate the template at {template_path}")
        raise typer.Exit(1)

    templateLoader = jinja2.FileSystemLoader(searchpath=".")
    templateEnv = jinja2.Environment(loader=templateLoader, trim_blocks=True, lstrip_blocks=True)
    template = templateEnv.get_template(str(template_path))

    try:
        rendered_tpl = template.render(**variables_dict, data=response)  # type: ignore[arg-type]
    except jinja2.TemplateSyntaxError as exc:
        console.print("[red]Syntax Error detected on the template")
        console.print(f"[yellow]  {exc}")
        raise typer.Exit(1) from exc

    except jinja2.UndefinedError as exc:
        console.print("[red]An error occured while rendering the jinja template")
        traceback = Traceback(show_locals=False)
        errors = identify_faulty_jinja_code(traceback=traceback)
        for frame, syntax in errors:
            console.print(f"[yellow]{frame.filename} on line {frame.lineno}\n")
            console.print(syntax)
        console.print("")
        console.print(traceback.trace.stacks[0].exc_value)
        raise typer.Exit(1) from exc

    print(rendered_tpl)


@app.command()
def run(
    script: Path,
    method: str = "run",
    debug: bool = False,
    config_file: str = typer.Option("infrahubctl.toml", envvar="INFRAHUBCTL_CONFIG"),
    branch: str = typer.Option("main", help="Branch on which to run the script."),
    concurrent: int = typer.Option(
        4,
        help="Maximum number of requets to execute at the same time.",
        envvar="INFRAHUBCTL_CONCURRENT_EXECUTION",
    ),
    timeout: int = typer.Option(60, help="Timeout in sec", envvar="INFRAHUBCTL_TIMEOUT"),
) -> None:
    """Execute a script."""

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.ERROR)
    logging.getLogger("httpcore").setLevel(logging.ERROR)

    log_level = "DEBUG" if debug else "INFO"
    FORMAT = "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
    log = logging.getLogger("infrahubctl")

    aiorun(
        _run(
            script=script,
            method=method,
            log=log,
            branch=branch,
            concurrent=concurrent,
            timeout=timeout,
        )
    )



import asyncio
from dataclasses import dataclass
from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, List, Optional, Tuple

from infrahub_sdk.node import InfrahubNode


@dataclass
class BatchTask:
    task: Callable[[Any], Awaitable[Any]]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    node: Optional[InfrahubNode] = None


async def execute_batch_task_in_pool(
    task: BatchTask, semaphore: asyncio.Semaphore
) -> Tuple[Optional[InfrahubNode], Any]:
    async with semaphore:
        return (task.node, await task.task(*task.args, **task.kwargs))


class InfrahubBatch:
    def __init__(
        self,
        semaphore: Optional[asyncio.Semaphore] = None,
        max_concurrent_execution: int = 5,
    ):
        self._tasks: List[BatchTask] = []
        self.semaphore = semaphore or asyncio.Semaphore(value=max_concurrent_execution)

    def add(
        self, *args: Any, task: Callable[[Any], Awaitable[Any]], node: Optional[InfrahubNode] = None, **kwargs: Any
    ) -> None:
        self._tasks.append(BatchTask(task=task, node=node, args=args, kwargs=kwargs))

    async def execute(self) -> AsyncGenerator:
        tasks = []

        for batch_task in self._tasks:
            tasks.append(asyncio.create_task(execute_batch_task_in_pool(task=batch_task, semaphore=self.semaphore)))

        for completed_task in asyncio.as_completed(tasks):
            yield await completed_task



import json
from typing import Any, Dict, Optional

import httpx
from pydantic import BaseSettings, Field

from infrahub_sdk.types import HTTPMethod
from infrahub_sdk.utils import generate_request_filename


class JSONPlayback(BaseSettings):
    directory: str = Field(default=".", description="Directory to read recorded files from")

    async def async_request(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        return self._read_request(url=url, method=method, headers=headers, payload=payload, timeout=timeout)

    def sync_request(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        return self._read_request(url=url, method=method, headers=headers, payload=payload, timeout=timeout)

    def _read_request(
        self,
        url: str,
        method: HTTPMethod,
        headers: Dict[str, Any],
        timeout: int,  # pylint: disable=unused-argument
        payload: Optional[Dict] = None,
    ) -> httpx.Response:
        content: Optional[bytes] = None
        if payload:
            content = str(json.dumps(payload)).encode("UTF-8")
        request = httpx.Request(method=method.value, url=url, headers=headers, content=content)

        filename = generate_request_filename(request)
        with open(f"{self.directory}/{filename}.json", "r", encoding="UTF-8") as fobj:
            data = json.load(fobj)

        response = httpx.Response(
            status_code=data["status_code"],
            content=data["response_content"],
            request=request,
        )
        return response

    class Config:
        env_prefix = "INFRAHUB_SDK_PLAYBACK_"
        case_sensitive = False



import logging
from asyncio import run as aiorun
from pathlib import Path
from typing import List, Optional

import typer
import yaml
from pydantic import BaseModel, ValidationError
from rich.console import Console
from rich.logging import RichHandler

import infrahub_ctl.config as config
from infrahub_ctl.client import initialize_client
from infrahub_sdk.utils import find_files

app = typer.Typer()


@app.callback()
def callback() -> None:
    """
    Manage the schema in a remote Infrahub instance.
    """


class SchemaFile(BaseModel):
    location: Path
    content: Optional[dict] = None
    valid: bool = True
    error_message: Optional[str] = None

    def load_content(self) -> None:
        try:
            self.content = yaml.safe_load(self.location.read_text())
        except yaml.YAMLError:
            self.error_message = "Invalid YAML/JSON file"
            self.valid = False


async def _load(schemas: List[Path], branch: str, log: logging.Logger) -> None:  # pylint: disable=unused-argument
    # pylint: disable=too-many-branches
    console = Console()

    schemas_data: List[SchemaFile] = []

    for schema in schemas:
        if schema.is_file():
            schema_file = SchemaFile(location=schema)
            schema_file.load_content()
            schemas_data.append(schema_file)
        elif schema.is_dir():
            files = find_files(extension=["yaml", "yml", "json"], directory=schema, recursive=True)
            for item in files:
                schema_file = SchemaFile(location=item)
                schema_file.load_content()
                schemas_data.append(schema_file)

    has_error = False
    for schema_file in schemas_data:
        if schema_file.valid:
            continue
        console.print(f"[red]{schema_file.error_message} ({schema_file.location})")
        has_error = True

    if has_error:
        raise typer.Exit(2)

    client = await initialize_client()

    # Valid data format of content
    for schema_file in schemas_data:
        try:
            client.schema.validate(schema_file.content)
        except ValidationError as exc:
            console.print(f"[red]Schema not valid, found '{len(exc.errors())}' error(s) in {schema_file.location}")
            has_error = True
            for error in exc.errors():
                loc_str = [str(item) for item in error["loc"]]
                console.print(f"  '{'/'.join(loc_str)}' | {error['msg']} ({error['type']})")

    if has_error:
        raise typer.Exit(2)

    _, errors = await client.schema.load(schemas=[item.content for item in schemas_data], branch=branch)

    if errors:
        console.print("[red]Unable to load the schema:")
        if "detail" in errors:
            for error in errors.get("detail"):
                loc_str = [str(item) for item in error["loc"][1:]]
                console.print(f"  '{'/'.join(loc_str)}' | {error['msg']} ({error['type']})")
        elif "error" in errors:
            console.print(f"  '{errors.get('error')}'")
        else:
            console.print(f"  '{errors}'")
    else:
        for schema_file in schemas_data:
            console.print(f"[green] schema '{schema_file.location}' loaded successfully!")


@app.command()
def load(
    schemas: List[Path],
    debug: bool = False,
    branch: str = typer.Option("main", help="Branch on which to load the schema."),
    config_file: str = typer.Option("infrahubctl.toml", envvar="INFRAHUBCTL_CONFIG"),
) -> None:
    """Load a schema file into Infrahub."""
    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.ERROR)
    logging.getLogger("httpcore").setLevel(logging.ERROR)

    log_level = "DEBUG" if debug else "INFO"
    FORMAT = "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
    log = logging.getLogger("infrahubctl")

    aiorun(_load(schemas=schemas, branch=branch, log=log))


@app.command()
def migrate() -> None:
    """Migrate the schema to the latest version. (Not Implemented Yet)"""
    print("Not implemented yet.")



QUERY_ALL_REPOSITORIES = """
query {
    CoreRepository {
        count
        edges {
            node {
                id
                name {
                    value
                }
                location {
                    value
                }
                commit {
                    value
                }
            }
        }
    }
}
"""


QUERY_ALL_BRANCHES = """
query {
    Branch {
        id
        name
        description
        origin_branch
        branched_from
        is_default
        is_data_only
    }
}
"""

QUERY_BRANCH = """
query GetBranch($branch_name: String!) {
    Branch(name: $branch_name) {
        id
        name
        description
        origin_branch
        branched_from
        is_default
        is_data_only
    }
}
"""


MUTATION_COMMIT_UPDATE = """
mutation ($repository_id: String!, $commit: String!) {
    CoreRepositoryUpdate(data: { id: $repository_id, commit: { value: $commit } }) {
        ok
        object {
            commit {
                value
            }
        }
    }
}
"""


QUERY_SCHEMA = """
    query {
        node_schema {
            name {
                value
            }
            kind {
                value
            }
            inherit_from {
                value
            }
            description {
                value
            }
            default_filter {
                value
            }
            attributes {
                name {
                    value
                }
                kind {
                    value
                }
                optional {
                    value
                }
                unique {
                    value
                }
                default_value {
                    value
                }
            }
            relationships {
                name {
                    value
                }
                peer {
                    value
                }
                identifier {
                    value
                }
                cardinality {
                    value
                }
                optional {
                    value
                }
            }
        }
    }
    """



from typing import Any, Dict, Optional

from pydantic import BaseSettings, Field, root_validator, validator

from infrahub_sdk.playback import JSONPlayback
from infrahub_sdk.recorder import JSONRecorder, Recorder, RecorderType
from infrahub_sdk.types import AsyncRequester, RequesterTransport, SyncRequester
from infrahub_sdk.utils import is_valid_url


class Config(BaseSettings):
    address: str = Field(
        default="http://localhost:8000",
        description="The URL to use when connecting to Infrahub.",
    )
    api_token: Optional[str] = Field(default=None, description="API token for authentication against Infrahub.")
    username: Optional[str] = Field(default=None, description="Username for accessing Infrahub", min_length=1)
    password: Optional[str] = Field(default=None, description="Password for accessing Infrahub", min_length=1)
    recorder: RecorderType = Field(
        default=RecorderType.NONE,
        description="Select builtin recorder for later replay.",
    )
    custom_recorder: Optional[Recorder] = Field(
        default=None,
        description="Provides a way to record responses from the Infrahub API",
    )
    requester: Optional[AsyncRequester] = None
    timeout: int = Field(default=10, description="Default connection timeout in seconds")
    transport: RequesterTransport = Field(
        default=RequesterTransport.HTTPX,
        description="Set an alternate transport using a predefined option",
    )
    sync_requester: Optional[SyncRequester] = None

    class Config:
        env_prefix = "INFRAHUB_SDK_"
        case_sensitive = False
        validate_assignment = True

    @root_validator(pre=True)
    @classmethod
    def validate_credentials_input(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        has_username = "username" in values
        has_password = "password" in values
        if has_username != has_password:
            raise ValueError("Both 'username' and 'password' needs to be set")
        return values

    @root_validator(pre=True)
    @classmethod
    def set_custom_recorder(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("recorder") == RecorderType.JSON and "custom_recorder" not in values:
            values["custom_recorder"] = JSONRecorder()
        return values

    @root_validator(pre=True)
    @classmethod
    def set_transport(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("transport") == RequesterTransport.JSON:
            playback = JSONPlayback()
            if "requester" not in values:
                values["requester"] = playback.async_request
            if "sync_requester" not in values:
                values["sync_requester"] = playback.sync_request

        return values

    @root_validator(pre=True)
    @classmethod
    def validate_mix_authentication_schemes(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("password") and values.get("api_token"):
            raise ValueError("Unable to combine password with token based authentication")
        return values

    @validator("address")
    @classmethod
    def validate_address(cls, value: str) -> str:
        if is_valid_url(value):
            return value

        raise ValueError("The configured address is not a valid url")

    @property
    def password_authentication(self) -> bool:
        if self.username:
            return True
        return False



from __future__ import annotations

from infrahub_sdk.batch import InfrahubBatch
from infrahub_sdk.branch import InfrahubBranchManager, InfrahubBranchManagerSync
from infrahub_sdk.client import InfrahubClient, InfrahubClientSync
from infrahub_sdk.config import Config
from infrahub_sdk.exceptions import (
    AuthenticationError,
    Error,
    FilterNotFound,
    GraphQLError,
    NodeNotFound,
    ServerNotReacheableError,
    ServerNotResponsiveError,
    ValidationError,
)
from infrahub_sdk.graphql import Mutation, Query
from infrahub_sdk.node import InfrahubNode, InfrahubNodeSync
from infrahub_sdk.schema import (
    AttributeSchema,
    InfrahubRepositoryConfig,
    InfrahubSchema,
    NodeSchema,
    RelationshipCardinality,
    RelationshipKind,
    RelationshipSchema,
    SchemaRoot,
)
from infrahub_sdk.store import NodeStore, NodeStoreSync
from infrahub_sdk.timestamp import Timestamp
from infrahub_sdk.uuidt import UUIDT, generate_uuid

__all__ = [
    "AttributeSchema",
    "AuthenticationError",
    "Config",
    "Error",
    "InfrahubBatch",
    "InfrahubBranchManager",
    "InfrahubBranchManagerSync",
    "InfrahubClient",
    "InfrahubClientSync",
    "InfrahubNode",
    "InfrahubNodeSync",
    "InfrahubRepositoryConfig",
    "InfrahubSchema",
    "FilterNotFound",
    "generate_uuid",
    "GraphQLError",
    "NodeNotFound",
    "NodeSchema",
    "Mutation",
    "NodeStore",
    "NodeStoreSync",
    "Query",
    "RelationshipSchema",
    "RelationshipCardinality",
    "RelationshipKind",
    "SchemaRoot",
    "ServerNotReacheableError",
    "ServerNotResponsiveError",
    "Timestamp",
    "UUIDT",
    "ValidationError",
]



import logging
import sys
from asyncio import run as aiorun
from datetime import datetime
from pathlib import Path
from typing import Dict, Generator, List, Optional, Union

import typer
from rich.console import Console
from rich.console import group as rich_group
from rich.panel import Panel
from rich.table import Table

import infrahub_ctl.config as config
from infrahub_ctl.client import initialize_client
from infrahub_ctl.utils import (
    calculate_time_diff,
    print_graphql_errors,
    render_action_rich,
)
from infrahub_sdk import Error, GraphQLError

app = typer.Typer()


DEFAULT_CONFIG_FILE = "infrahubctl.toml"
ENVVAR_CONFIG_FILE = "INFRAHUBCTL_CONFIG"


@app.callback()
def callback() -> None:
    """
    Manage the branches in a remote Infrahub instance.

    List, create, merge, rebase ..
    """


async def _list() -> None:
    client = await initialize_client()

    console = Console()

    try:
        branches = await client.branch.all()
    except GraphQLError as exc:
        print_graphql_errors(console, exc.errors)
        sys.exit(1)
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    table = Table(title="List of all branches")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Description")
    table.add_column("Origin Branch")
    table.add_column("Branched From")
    table.add_column("Is Data Only")
    table.add_column("Is Default")

    # identify the default branch and always print it first
    default_branch = [branch for branch in branches.values() if branch.is_default][0]
    table.add_row(
        default_branch.name,
        default_branch.description or " - ",
        default_branch.origin_branch,
        f"{default_branch.branched_from} ({calculate_time_diff(default_branch.branched_from)})",
        "[green]True" if default_branch.is_data_only else "[#FF7F50]False",
        "[green]True" if default_branch.is_default else "[#FF7F50]False",
    )

    for branch in branches.values():
        if branch.is_default:
            continue

        table.add_row(
            branch.name,
            branch.description or " - ",
            branch.origin_branch,
            f"{branch.branched_from} ({calculate_time_diff(branch.branched_from)})",
            "[green]True" if branch.is_data_only else "[#FF7F50]False",
            "[green]True" if branch.is_default else "[#FF7F50]False",
        )

    console.print(table)


@app.command("list")
def list_branch(
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """List all existing branches."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_list())


async def _create(branch_name: str, description: str, data_only: bool) -> None:
    console = Console()

    client = await initialize_client()

    try:
        branch = await client.branch.create(branch_name=branch_name, description=description, data_only=data_only)
    except GraphQLError as exc:
        print_graphql_errors(console, exc.errors)
        sys.exit(1)
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    console.print(f"Branch '{branch_name}' created successfully ({branch.id}).")


@app.command()
def create(
    branch_name: str,
    description: str = typer.Argument(""),
    data_only: bool = True,
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """Create a new branch."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_create(branch_name=branch_name, description=description, data_only=data_only))


async def _delete(branch_name: str) -> None:
    console = Console()

    client = await initialize_client()

    try:
        await client.branch.delete(branch_name=branch_name)
    except GraphQLError as exc:
        print_graphql_errors(console, exc.errors)
        sys.exit(1)
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    console.print(f"Branch '{branch_name}' deleted successfully.")


@app.command()
def delete(
    branch_name: str,
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """Delete a branch."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_delete(branch_name=branch_name))


async def _rebase(branch_name: str) -> None:
    console = Console()

    client = await initialize_client()

    try:
        await client.branch.rebase(branch_name=branch_name)
    except GraphQLError as exc:
        print_graphql_errors(console, exc.errors)
        sys.exit(1)
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    console.print(f"Branch '{branch_name}' rebased successfully.")


@app.command()
def rebase(
    branch_name: str,
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """Rebase a Branch with main."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_rebase(branch_name=branch_name))


async def _merge(branch_name: str) -> None:
    console = Console()

    client = await initialize_client()

    try:
        await client.branch.merge(branch_name=branch_name)
    except GraphQLError as exc:
        print_graphql_errors(console, exc.errors)
        sys.exit(1)
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    console.print(f"Branch '{branch_name}' merged successfully.")


@app.command()
def merge(
    branch_name: str,
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """Merge a Branch with main."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_merge(branch_name=branch_name))


async def _validate(branch_name: str) -> None:
    console = Console()

    client = await initialize_client()

    try:
        await client.branch.validate(branch_name=branch_name)
    except GraphQLError as exc:
        print_graphql_errors(console, exc.errors)
        sys.exit(1)
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    console.print(f"Branch '{branch_name}' is valid.")


@app.command()
def validate(
    branch_name: str,
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """Validate if a branch has some conflict and is passing all the tests (NOT IMPLEMENTED YET)."""

    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    aiorun(_validate(branch_name=branch_name))


@rich_group()
def node_panel_generator(nodes: List[Dict]) -> Generator:
    for node in nodes:
        lines = []

        for attr in node["attributes"]:
            attr_header = "{:10}{:10}  {:<20}".format(
                "",
                render_action_rich(attr["action"]),
                attr["name"],
            )
            lines.append(attr_header)

            for prop in attr["properties"]:
                clean_prop_name = prop["type"].replace("HAS_", "").replace("IS_", "").lower()

                property_string = "{:>20}[magenta]{:>20}[blue]{:>4}[magenta]{:20}[green]{:>20}[/green]{:>20}".format(
                    clean_prop_name,
                    f"{prop['value']['previous']}",
                    " >> ",
                    f"{prop['value']['new']}",
                    f"{prop['changed_at']}",
                    f"({calculate_time_diff(prop['changed_at'])})",
                )

            lines.append(property_string)

        for rel in node["relationships"]:
            rel_header = "{:10}{:10}  {:<20} Node {:>20}{:>20} {:>20}".format(
                "",
                render_action_rich(rel["action"]),
                rel["name"],
                rel["peer"]["kind"],
                rel["peer"]["display_label"],
                rel["peer"]["id"],
            )

            lines.append(rel_header)

            for prop in rel["properties"]:
                clean_prop_name = prop["type"].replace("HAS_", "").replace("IS_", "").lower()

                property_string = "{:>20}[magenta]{:>20}[blue]{:>4}[magenta]{:20}[green]{:>20}[/green]{:>20}".format(
                    clean_prop_name,
                    f"{prop['value']['previous']}",
                    " >> ",
                    f"{prop['value']['new']}",
                    f"{prop['changed_at']}",
                    f"({calculate_time_diff(prop['changed_at'])})",
                )

            lines.append(property_string)

        yield Panel(
            "\n".join(lines),
            title=f"Node {node['kind']} {node['display_label']} ({node['id']})",
            title_align="left",
        )


async def _diff(
    branch_name: str,
    time_from: Union[str, datetime],
    time_to: Union[str, datetime],
    branch_only: bool,
) -> None:
    console = Console()

    client = await initialize_client()

    try:
        response = await client.branch.diff_data(
            branch_name=branch_name,
            branch_only=branch_only,
            time_from=time_from,
            time_to=time_to,
        )
    except Error as exc:
        console.print(f"[red]{exc.message}")
        sys.exit(1)

    for branch, nodes in response.json().items():
        console.print(
            Panel(
                node_panel_generator(nodes),
                title=f"Branch: [green]{branch}",
                title_align="left",
            )
        )


@app.command()
def diff(
    branch_name: str,
    time_from: Optional[datetime] = typer.Option(
        None,
        "--from",
        help="Start Time used to calculate the Diff, Default: from the start of the branch",
    ),
    time_to: Optional[datetime] = typer.Option(None, "--to", help="End Time used to calculate the Diff, Default: now"),
    branch_only: bool = True,
    config_file: Path = typer.Option(DEFAULT_CONFIG_FILE, envvar=ENVVAR_CONFIG_FILE),
) -> None:
    """Show the differences between a Branch and main."""
    if not config.SETTINGS:
        config.load_and_exit(config_file=config_file)

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    aiorun(
        _diff(
            branch_name=branch_name,
            time_from=time_from or "",
            time_to=time_to or "",
            branch_only=branch_only,
        )
    )



"""Config Clas."""
from pathlib import Path
from typing import Any, Dict, Optional, Union

import toml
import typer
from pydantic import BaseSettings, ValidationError, root_validator

DEFAULT_CONFIG_FILE = "infrahubctl.toml"
ENVVAR_CONFIG_FILE = "INFRAHUBCTL_CONFIG"
INFRAHUB_REPO_CONFIG_FILE = ".infrahub.yml"


class Settings(BaseSettings):
    """Main Settings Class for the project."""

    server_address: str = "http://localhost:8000"
    api_token: Optional[str] = None
    default_branch: str = "main"

    class Config:
        """Additional parameters to automatically map environment variable to some settings."""

        fields = {
            "server_address": {"env": "INFRAHUB_ADDRESS"},
            "api_token": {"env": "INFRAHUB_API_TOKEN"},
            "default_branch": {"env": "INFRAHUB_DEFAULT_BRANCH"},
        }

    @root_validator
    def cleanup_server_address(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # pylint: disable=no-self-argument
        values["server_address"] = values["server_address"].rstrip("/")
        return values


def load(
    config_file: Union[str, Path] = "infrahubctl.toml",
    config_data: Optional[dict] = None,
) -> None:
    """Load configuration.

    Configuration is loaded from a config file in toml format that contains the settings,
    or from a dictionary of those settings passed in as "config_data"
    """
    global SETTINGS  # pylint: disable=global-statement

    if config_data:
        SETTINGS = Settings(**config_data)
        return

    if not isinstance(config_file, Path):
        config_file = Path(config_file)

    if config_file.is_file():
        config_string = config_file.read_text(encoding="utf-8")
        config_tmp = toml.loads(config_string)

        SETTINGS = Settings(**config_tmp)
        return

    SETTINGS = Settings()


def load_and_exit(
    config_file: Union[str, Path] = "infrahubctl.toml",
    config_data: Optional[dict] = None,
) -> None:
    """Calls load, but wraps it in a try except block.

    This is done to handle a ValidationErorr which is raised when settings are specified but invalid.
    In such cases, a message is printed to the screen indicating the settings which don't pass validation.

    Args:
        config_file_name (str, optional): [description]. Defaults to "pyprojectctl.toml".
        config_data (dict, optional): [description]. Defaults to None.
    """
    try:
        load(config_file=config_file, config_data=config_data)
    except ValidationError as exc:
        print(f"Configuration not valid, found {len(exc.errors())} error(s)")
        for error in exc.errors():
            loc_str = [str(item) for item in error["loc"]]
            print(f"  {'/'.join(loc_str)} | {error['msg']} ({error['type']})")
        raise typer.Abort()


SETTINGS = Settings()



from typing import Any

import infrahub_ctl.config as config
from infrahub_sdk import InfrahubClient, InfrahubClientSync
from infrahub_sdk.config import Config


async def initialize_client(**kwargs: Any) -> InfrahubClient:
    client_config = {}

    if config.SETTINGS.api_token:
        client_config["api_token"] = config.SETTINGS.api_token

    timeout = kwargs.pop("timeout", None)
    if timeout:
        client_config["timeout"] = timeout

    client = await InfrahubClient.init(
        address=config.SETTINGS.server_address, config=Config(**client_config), insert_tracker=True, **kwargs
    )

    return client


def initialize_client_sync(**kwargs: Any) -> InfrahubClientSync:
    client_config = {}

    if config.SETTINGS.api_token:
        client_config["api_token"] = config.SETTINGS.api_token

    client = InfrahubClientSync.init(
        address=config.SETTINGS.server_address, config=Config(**client_config), insert_tracker=True, **kwargs
    )

    return client



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Dict, Optional

from infrahub_sdk.exceptions import NodeNotFound

if TYPE_CHECKING:
    from infrahub_sdk.node import InfrahubNode, InfrahubNodeSync


class NodeStoreBase:
    """Internal Store for InfrahubNode objects.

    Often while creating a lot of new objects,
    we need to save them in order to reuse them laterto associate them with another node for example.
    """

    def __init__(self) -> None:
        self._store: Dict[str, Dict] = defaultdict(dict)

    def _set(self, key: str, node) -> None:  # type: ignore[no-untyped-def]
        if "InfrahubNode" not in node.__class__.__name__:
            raise TypeError(f"'node' must be of type InfrahubNode, not {type(node)!r}")

        node_kind = node._schema.kind
        self._store[node_kind][key] = node  # type: ignore[attr-defined]

    def _get(self, key: str, kind: Optional[str] = None, raise_when_missing: bool = True):  # type: ignore[no-untyped-def]
        if kind and kind not in self._store and key not in self._store[kind]:  # type: ignore[attr-defined]
            if not raise_when_missing:
                return None
            raise NodeNotFound(
                branch_name="n/a",
                node_type=kind,
                identifier={"key": [key]},
                message="Unable to find the node in the Store",
            )

        if kind and kind in self._store and key in self._store[kind]:  # type: ignore[attr-defined]
            return self._store[kind][key]  # type: ignore[attr-defined]

        for _, item in self._store.items():  # type: ignore[attr-defined]
            if key in item:
                return item[key]

        if not raise_when_missing:
            return None
        raise NodeNotFound(
            branch_name="n/a",
            node_type="n/a",
            identifier={"key": [key]},
            message=f"Unable to find the node {key!r} in the Store",
        )


class NodeStore(NodeStoreBase):
    def get(self, key: str, kind: Optional[str] = None, raise_when_missing: bool = True) -> Optional[InfrahubNode]:
        return self._get(key=key, kind=kind, raise_when_missing=raise_when_missing)

    def set(self, key: str, node: InfrahubNode) -> None:
        return self._set(key=key, node=node)


class NodeStoreSync(NodeStoreBase):
    def get(self, key: str, kind: Optional[str] = None, raise_when_missing: bool = True) -> Optional[InfrahubNodeSync]:
        return self._get(key=key, kind=kind, raise_when_missing=raise_when_missing)

    def set(self, key: str, node: InfrahubNodeSync) -> None:
        return self._set(key=key, node=node)
